#!/usr/bin/python3
########################################################################################
# __utils__.py - Profile-generation utility module.                                    #
#                                                                                      #
# Author(s): Phil Sandwell, Ben Winchester                                             #
# Copyright: Phil Sandwell, 2021                                                       #
# Date created: 11/08/2021                                                             #
# License: Open source                                                                 #
# Most recent update: 14/07/2021                                                       #
#                                                                                      #
# Additional credits:                                                                  #
#     Iain Staffell, Stefan Pfenninger & Scot Wheeler                                  #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
__utils__.py - The profile-generation utility module for CLOVER.

This module fetches profiles from renewables.ninja, parses them and saves them for use
locally within CLOVER. The profiles that are fetched are determined by the information
that is passed in to the module.

"""

from argparse import Namespace
import dataclasses
from logging import Logger

from typing import Any, Dict, List, Optional, Union

from ..__utils__ import (
    AuxiliaryHeaterType,
    BColours,
    CleanWaterMode,
    EXCHANGER,
    InputFileError,
    NAME,
    RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING,
    OperatingMode,
    ResourceType,
    Scenario,
)

from ..conversion.conversion import Converter
from ..generation.solar import HybridPVTPanel, PVPanel
from .diesel import DieselGenerator, DieselWaterHeater
from .exchanger import Exchanger
from .storage_utils import Battery, CleanWaterTank, HotWaterTank
from .transmission import Transmitter

__all__ = (
    "check_scenario",
    "determine_available_converters",
    "Minigrid",
)

# AC-to-AC:
#   Keyword used for parsing ac-to-ac conversion parameters.
AC_TO_AC: str = "ac_to_ac"

# AC-to-DC:
#   Keyword used for parsing ac-to-dc conversion parameters.
AC_TO_DC: str = "ac_to_dc"

# Conversion:
#   Keyword used for parsing grid AC-DC conversion parameters.
CONVERSION: str = "conversion"

# DC-to-AC:
#   Keyword used for parsing dc-to-ac conversion parameters.
DC_TO_AC: str = "dc_to_ac"

# DC-to-DC:
#   Keyword used for parsing dc-to-dc conversion parameters.
DC_TO_DC: str = "dc_to_dc"

# Resource Type:
#   Used for parsing resource-type information.
RESOURCE_TYPE: str = "resource_type"


@dataclasses.dataclass
class Minigrid:
    """
    Represents an energy system.

    .. attribute:: ac_to_ac_conversion_efficiency
        The conversion efficiency from AC to AC.

    .. attribute:: ac_to_dc_conversion_efficiency
        The conversion efficiency from AC to DC.

    .. attribute:: ac_transmission_efficiency
        The AC transmission efficiency, if applicable.

    .. attribute:: battery
        The battery being modelled, if applicable.

    .. attribute:: buffer_tank
        The buffer tank being modelled, if applicable. This tank contains the buffer
        solution, be it HTF or feedwater, which is heated by PV-T before being fed into
        the desalination plant.

    .. attribute:: clean_water_tank
        The clean-water tank being modelled, if applicable.

    .. attribute:: dc_to_ac_conversion_efficiency
        The conversion efficiency from DC to AC.

    .. attribute:: dc_to_dc_conversion_efficiency
        The conversion efficiency from DC to DC.

    .. attribute:: dc_transmission_efficiency
        The DC transmission efficiency, if applicable.

    .. attribute:: diesel_generator
        The diesel backup generator associated with the minigrid system.

    .. attribute:: diesel_water_heater
        The diesel water heater associated with the minigrid system.

    .. attribute:: electric_water_heater
        The electric water heater associated with the minigrid system.

    .. attribute:: heat_exchanger
        The heat exchanger associated with the minigrid system.

    .. attribute:: hot_water_tank
        The hot-water tank being modelled, if applicable.

    .. attribute:: pv_panel
        The PV panel being considered.

    .. attribute:: pvt_panel
        The PV-T panel being considered, if applicable.

    .. attribute:: water_pump
        The water pump associated with the energy system, as a :class:`Transmitter`
        instance.

    """

    ac_to_ac_conversion_efficiency: Optional[float]
    ac_to_dc_conversion_efficiency: Optional[float]
    ac_transmission_efficiency: Optional[float]
    battery: Optional[Battery]
    buffer_tank: Optional[HotWaterTank]
    clean_water_tank: Optional[CleanWaterTank]
    dc_to_ac_conversion_efficiency: Optional[float]
    dc_to_dc_conversion_efficiency: Optional[float]
    dc_transmission_efficiency: Optional[float]
    diesel_generator: Optional[DieselGenerator]
    diesel_water_heater: Optional[DieselWaterHeater]
    electric_water_heater: Optional[Converter]
    heat_exchanger: Optional[Exchanger]
    hot_water_tank: Optional[HotWaterTank]
    pv_panel: PVPanel
    pvt_panel: Optional[HybridPVTPanel]
    water_pump: Optional[Transmitter]

    @classmethod
    def from_dict(  # pylint: disable=too-many-locals
        cls,
        diesel_generator: DieselGenerator,
        diesel_water_heater: Optional[DieselWaterHeater],
        electric_water_heater: Optional[Converter],
        minigrid_inputs: Dict[str, Any],
        pv_panel: PVPanel,
        pvt_panel: Optional[HybridPVTPanel],
        battery_inputs: Optional[List[Dict[str, Any]]] = None,
        exchanger_inputs: Optional[List[Dict[str, Any]]] = None,
        tank_inputs: Optional[List[Dict[str, Any]]] = None,
        water_pump: Optional[Transmitter] = None,
    ) -> Any:
        """
        Returns a :class:`Minigrid` instance based on the inputs provided.

        Inputs:
            - diesel_generator:
                The diesel backup generator to use for the run.
            - diesel_water_heater:
                The diesel water heater associated with the minigrid system, if
                appropriate.
            - electric_water_heater:
                The electric water heater associated with the minigrid system, if
                appropriate.
            - minigrid_inputs:
                The inputs for the minigrid/energy system, extracted from the input
                file.
            - pv_panel:
                The :class:`PVPanel` instance to use for the run.
            - pvt_panel:
                The :class:`HybridPVTPanel` instance to use for the run, if appropriate.
            - battery_inputs:
                The battery input information.
            - exchanger_inputs:
                The heat-exchanger input information.
            - tank_inputs:
                The tank input information.
            - water_pump:
                The :class`Transmitter` corresponding to the water pump associated with
                the system.

        Outputs:
            - A :class:`Minigrid` instance based on the inputs provided.

        """

        # Parse the battery information.
        if battery_inputs is not None:
            batteries = {
                entry[NAME]: Battery.from_dict(entry) for entry in battery_inputs
            }
        else:
            batteries = {}

        # Parse the heat-exchanger information.
        if exchanger_inputs is not None:
            exchangers = {
                entry[NAME]: Exchanger.from_dict(entry) for entry in exchanger_inputs
            }
        else:
            exchangers = {}

        tanks: Dict[str, Union[CleanWaterTank, HotWaterTank]] = {}
        # Parse the tank information.
        if tank_inputs is not None:
            for entry in tank_inputs:
                if (
                    RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING[entry[RESOURCE_TYPE]]
                    == ResourceType.CLEAN_WATER
                ):
                    try:
                        tanks[entry[NAME]] = CleanWaterTank.from_dict(entry)
                    except KeyError as e:
                        raise InputFileError(
                            "tank inputs",
                            f"Error parsing clean-water tank {entry['name']}: {str(e)}",
                        ) from None
                elif (
                    RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING[entry[RESOURCE_TYPE]]
                    == ResourceType.HOT_CLEAN_WATER
                ):
                    try:
                        tanks[entry[NAME]] = HotWaterTank.from_dict(entry)
                    except KeyError as e:
                        raise InputFileError(
                            "tank inputs",
                            f"Error parsing hot-water tank {entry['name']}: {str(e)}",
                        ) from None
                else:
                    raise InputFileError(
                        "tank inputs",
                        f"The tank '{entry['name']}' uses an unknown resource type: "
                        + f"{entry[RESOURCE_TYPE]}",
                    )
        else:
            tanks = {}

        # Determine the various tanks being considered.
        if "buffer_tank" in minigrid_inputs:
            buffer_tank: Optional[Union[CleanWaterTank, HotWaterTank]] = tanks[
                minigrid_inputs["buffer_tank"]
            ]
            if not isinstance(buffer_tank, HotWaterTank):
                raise InputFileError(
                    "energy system inputs",
                    "The buffer tank selected must be a hot-water tank.",
                )
        else:
            buffer_tank = None

        if "clean_water_tank" in minigrid_inputs:
            clean_water_tank: Optional[Union[CleanWaterTank, HotWaterTank]] = tanks[
                minigrid_inputs["clean_water_tank"]
            ]
            if not isinstance(clean_water_tank, CleanWaterTank):
                raise InputFileError(
                    "energy system inputs",
                    "The clean-water tank selected must be a clean-water tank.",
                )
        else:
            clean_water_tank = None

        if "hot_water_tank" in minigrid_inputs:
            hot_water_tank: Optional[Union[CleanWaterTank, HotWaterTank]] = tanks[
                minigrid_inputs["hot_water_tank"]
            ]
            if not isinstance(hot_water_tank, HotWaterTank):
                raise InputFileError(
                    "energy system inputs",
                    "The hot-water tank selected must be a hot-water tank.",
                )
        else:
            hot_water_tank = None

        # Return the minigrid instance.
        return cls(
            minigrid_inputs[CONVERSION][AC_TO_AC]
            if AC_TO_AC in minigrid_inputs[CONVERSION]
            else None,
            minigrid_inputs[CONVERSION][AC_TO_DC]
            if AC_TO_DC in minigrid_inputs[CONVERSION]
            else None,
            minigrid_inputs["ac_transmission_efficiency"]
            if "ac_transmission_efficiency" in minigrid_inputs
            else None,
            batteries[minigrid_inputs["battery"]]
            if "battery" in minigrid_inputs
            else None,
            buffer_tank,
            clean_water_tank,
            minigrid_inputs[CONVERSION][DC_TO_AC]
            if DC_TO_AC in minigrid_inputs[CONVERSION]
            else None,
            minigrid_inputs[CONVERSION][DC_TO_DC]
            if DC_TO_DC in minigrid_inputs[CONVERSION]
            else None,
            minigrid_inputs["dc_transmission_efficiency"]
            if "dc_transmission_efficiency" in minigrid_inputs
            else None,
            diesel_generator,
            diesel_water_heater,
            electric_water_heater,
            exchangers[minigrid_inputs[EXCHANGER]]
            if EXCHANGER in minigrid_inputs
            else None,
            hot_water_tank,
            pv_panel,
            pvt_panel,
            water_pump,
        )


def check_scenario(
    logger: Logger,
    minigrid: Minigrid,
    operating_mode: OperatingMode,
    parsed_args: Namespace,
    scenario: Scenario,
) -> None:
    """
    Check that the scenario inputs match up correctly.

    Inputs:
        - logger:
            The logger to use for the run.
        - minigrid:
            The minigrid being simulated.
        - operating_mode:
            The operating mode being carried out.
        - parsed_args:
            The parsed command-line arguments.
        - scenario:
            The current scenario beign modelled.

    """

    if scenario.desalination_scenario is not None and minigrid.clean_water_tank is None:
        raise InputFileError(
            "energy system inputs",
            "No clean-water tank was provided despite there needing to be a tank "
            "specified for dealing with clean-water demands.",
        )

    if operating_mode == OperatingMode.SIMULATION:
        if (scenario.pv and parsed_args.pv_system_size is None) or (
            not scenario.pv and parsed_args.pv_system_size is not None
        ):
            raise InputFileError(
                "scenario",
                "PV mode in the scenario file must match the command-line usage.",
            )
        if (
            parsed_args.clean_water_pvt_system_size is not None
            and (scenario.desalination_scenario is None)
        ) or (
            parsed_args.clean_water_pvt_system_size is None
            and (
                scenario.desalination_scenario is not None
                and scenario.desalination_scenario.clean_water_scenario.mode
                == CleanWaterMode.THERMAL_ONLY
            )
        ):
            logger.error(
                "%sPV-T mode and available resources in the scenario file must match "
                "the command-line usage. Check the clean-water and PV-T scenario "
                "specification.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "scenario",
                "Mismatch between command-line usage and in-file usage.",
            )
        if (
            parsed_args.hot_water_pvt_system_size is not None
            and (scenario.hot_water_scenario is None)
        ) or (
            parsed_args.hot_water_pvt_system_size is None
            and (scenario.hot_water_scenario is not None)
        ):
            logger.error(
                "%sPV-T mode in the scenario file must match the command-line usage. "
                "Check the hot-water and PV-T scenario specification.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "scenario",
                "Mismatch between command-line usage and in-file usage.",
            )
        if (
            scenario.pv_t
            and scenario.desalination_scenario is None
            and scenario.hot_water_scenario is None
        ) or (
            not scenario.pv_t
            and scenario.desalination_scenario is not None
            and scenario.hot_water_scenario is not None
        ):
            logger.error(
                "%sDesalination or hot-water scenario usage does not match the "
                "system's PV-T panel inclusion.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "scenario",
                "The PV-T mode does not match the hot-water or desalination scenarios.",
            )
        if (scenario.battery and parsed_args.storage_size is None) or (
            not scenario.battery and parsed_args.storage_size is not None
        ):
            raise InputFileError(
                "scenario",
                "Battery mode in the scenario file must match the command-line usage.",
            )


def determine_available_converters(
    converters: Dict[str, Converter],
    logger: Logger,
    minigrid: Minigrid,
    scenario: Scenario,
) -> List[Converter]:
    """
    Determines the available :class:`Converter` instances based on the :class:`Scenario`

    Inputs:
        - converters:
            The :class:`Converter` instances defined, parsed from the conversion inputs
            file.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - minigrid:
            The :class:`Minigrid` to use for the run.
        - scenario:
            The :class:`Scenario` to use for the run.

    Outputs:
        - A `list` of :class:`Converter` instances available to the system.

    """

    available_converters: List[Converter] = []

    if scenario.desalination_scenario is None and scenario.hot_water_scenario is None:
        return available_converters

    # Determine the available converters from the scenarios file.
    if scenario.desalination_scenario is not None:
        # Process the clean-water converters.
        for entry in scenario.desalination_scenario.clean_water_scenario.sources:
            try:
                available_converters.append(converters[entry])
            except KeyError:
                logger.error(
                    "%sUnknown clean-water source specified in the scenario file: %s%s",
                    BColours.fail,
                    entry,
                    BColours.endc,
                )
                raise InputFileError(
                    "desalination scenario",
                    f"{BColours.fail}Unknown clean-water source(s) in the scenario "
                    + f"file: {entry}{BColours.endc}",
                ) from None

        # Process the feedwater sources.
        for entry in scenario.desalination_scenario.unclean_water_sources:
            try:
                available_converters.append(converters[entry])
            except KeyError:
                logger.error(
                    "%sUnknown unclean-water source specified in the scenario file: %s"
                    "%s",
                    BColours.fail,
                    entry,
                    BColours.endc,
                )
                raise InputFileError(
                    "desalination scenario",
                    f"{BColours.fail}Unknown unclean-water source in the scenario "
                    + f"file: {entry}{BColours.endc}",
                ) from None

    if scenario.hot_water_scenario is not None:
        # Process the hot-water converters.
        for entry in scenario.hot_water_scenario.conventional_sources:
            try:
                available_converters.append(converters[entry])
            except KeyError:
                logger.error(
                    "%sUnknown conventional hot-water source specified in the "
                    "hot-water scenario file: %s%s",
                    BColours.fail,
                    entry,
                    BColours.endc,
                )
                raise InputFileError(
                    "hot-water scenario",
                    f"{BColours.fail}Unknown conventional hot-water source(s) in the "
                    + f"hot-water scenario file: {entry}{BColours.endc}",
                ) from None

        if scenario.hot_water_scenario.auxiliary_heater == AuxiliaryHeaterType.ELECTRIC:
            if minigrid.electric_water_heater is None:
                logger.error(
                    "%sAuxiliary heating method of electric heating specified despite "
                    "no electric water heater being selected in the energy-system "
                    "inputs.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InputFileError(
                    "energy system inputs OR hot-water scenario",
                    "Mismatch between electric water heating scenario.",
                )
            available_converters.append(converters[minigrid.electric_water_heater.name])

    return available_converters



#!/usr/bin/python3
########################################################################################
# exchanger.py - Heat-exchanger module for CLOVER.                                     #
#                                                                                      #
# Authors: Phil Sandwell, Ben Winchester                                               #
# Copyright: Phil Sandwell, 2018                                                       #
# Date created: 25/10/2021                                                             #
# License: Open source                                                                 #
#                                                                                      #
# For more information, please email:                                                  #
#   philip.sandwell@gmail.com                                                          #
########################################################################################
"""
The exchanger module for CLOVER.

This module represents the heat exchanger within CLOVER energy systems. The heat
exchanger is usually located within the hot-water tank, but CLOVER may access the
heat-exchanger related information from anywhere within CLOVER.

"""

import dataclasses

from typing import Any, Dict

from ..__utils__ import NAME

__all__ = ("Exchanger",)


# Efficiency:
#   Keyword used to parse the efficiency of the heat exchanger.
EFFICIENCY: str = "efficiency"


@dataclasses.dataclass
class Exchanger:
    """
    Represents a physical heat exchanger within a hot-water tank.

    .. attribute:: efficiency
        The efficiency of the heat exchanger, defined between 0 and 1.

    """

    efficiency: float
    name: str

    def __repr__(self) -> str:
        """
        Returns a nice representation of the heat exchanger.

        :return:
            A `str` giving a nice representation of the heat exchanger.

        """

        return f"Exchanger(name={self.name}, efficiency={self.efficiency})"

    @classmethod
    def from_dict(cls, exchanger_inputs: Dict[str, Any]) -> Any:
        """
        Instantiate a :class:`Exchanger` based on the input information.

        Inputs:
            - exchanger_inputs:
                The exchanger input informaiton.

        Outputs:
            - A :class:`Exchanger` instance.

        """

        return cls(exchanger_inputs[EFFICIENCY], exchanger_inputs[NAME])



#!/usr/bin/env python
########################################################################################
# clover_new_location.py - Entry point for creating new locations.                     #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 16/09/2021                                                             #
# License: Open source                                                                 #
########################################################################################
"""
clover_new_location.py - Entry point for creating new locations.

CLOVER works on the basis of locations, and, as such, the ability to create new
locations is integral to the operation of CLOVER.

"""

import sys

from .new_location import main as new_location_main


def main() -> None:
    """
    Main function of the new-location entry-point script.

    """

    new_location_main(sys.argv[1:])



#!/usr/bin/python3
########################################################################################
# weather.py - Weather generation module  .                                            #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 11/08/2021                                                             #
# License: Open source                                                                 #
# Most recent update: 11/08/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
weather.py - The weather-profile-generation module for CLOVER.

This module fetches weather profiles from renewables.ninja, parses them and saves them
for use locally within CLOVER.

"""

import enum

from typing import Any, Dict

import pandas as pd  # pylint: disable=import-error


from .__utils__ import BaseRenewablesNinjaThread, total_profile_output
from ..__utils__ import Location

__all__ = (
    "total_weather_output",
    "WeatherCondition",
    "WeatherDataThread",
    "WEATHER_LOGGER_NAME",
)


# Weather logger name:
#   The name to use for the weather logger.
WEATHER_LOGGER_NAME = "weather_generation"


class WeatherCondition(enum.Enum):
    """
    Stores weather condition information for extracting data from renewables ninja.

    - CLOUD_COVER:
        Denotes the cloud-cover fraction.

    - IRRADIANCE:
        Denotes the surface irradiance.

    - PRECIPITATION:
        Denotes the surface precipitation.

    - TEMPERATURE:
        Denotes the surface temperature in degrees Celcius.

    """

    CLOUD_COVER = "cloud_cover"
    IRRADIANCE = "surface_radiation"
    PRECIPITATION = "precipitation"
    TEMPERATURE = "temperature"


class WeatherDataThread(
    BaseRenewablesNinjaThread, profile_name="weather", profile_key="weather"
):
    """
    Class to use when calling the weather data thread.

    """

    def __init__(
        self,
        auto_generated_files_directory: str,
        generation_inputs: Dict[str, Any],
        location: Location,
        logger_name: str,
        regenerate: bool,
        sleep_multiplier: int = 1,
        verbose: bool = False,
    ):
        """
        Instantiate a :class:`WeatherDataThread` instance.

        The weather variables that can be fetched are:
            - var_t2m:
                Surface temperature in degrees Celcius.
            - var_prectotland:
                Precipitation in mm/hour.
            - var_precsnoland
                Snowfall in mm/hour.
            - var_snomas:
                The mass of snow per land area in mm/m^2.
            - var_rhoa:
                The air density, at ground level, measured in kg/m^3.
            - var_swgdn:
                The solar irradiance at ground level, measured in W/m^2.
            - var_swtdn:
                The solar irradiance at the top of the atmosphere, measured in W/m^2.
            - var_cldtot:
                The cloud cover fraction, defined between 0 and 1.

        """

        # Add the additional parameters which are need when calling the weather data.
        renewables_ninja_params = {
            "dataset": "merra2",
            "lat": float(location.latitude),
            "lon": float(location.longitude),
            "local_time": "false",
            "header": "true",
            "var_t2m": "true",
            "var_prectotland": "true",
            "var_swgdn": "true",
            "var_cldtot": "true",
        }
        super().__init__(
            auto_generated_files_directory,
            generation_inputs,
            location,
            logger_name,
            regenerate,
            sleep_multiplier,
            verbose,
            renewables_ninja_params=renewables_ninja_params,
        )


def total_weather_output(*args, **kwargs) -> pd.DataFrame:
    """
    Wrapper function to wrap the total weather output.

    """

    return total_profile_output(*args, **kwargs, profile_name="weather")



#!/usr/bin/python3
########################################################################################
# solar.py - Solar generation module  .                                                #
#                                                                                      #
# Author: Phil Sandwell                                                                #
# Copyright: Phil Sandwell, 2021                                                       #
# License: Open source                                                                 #
# Most recent update: 14/07/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
solar.py - The solar-profile-generation module for CLOVER.

This module fetches solar profiles from renewables.ninja, parses them and saves them
for use locally within CLOVER.

"""

import enum

from logging import Logger
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd  # pylint: disable=import-error

# from sklearn.linear_model._coordinate_descent import Lasso

from ..__utils__ import (
    BColours,
    InputFileError,
    Location,
    NAME,
    ProgrammerJudgementFault,
)
from .__utils__ import BaseRenewablesNinjaThread, SolarDataType, total_profile_output

__all__ = (
    "HybridPVTPanel",
    "PVPanel",
    "SolarDataThread",
    "SolarDataType",
    "SolarPanelType",
    "solar_degradation",
    "SOLAR_LOGGER_NAME",
    "total_solar_output",
)


# Default PV unit:
#   The default PV unit size to use, measured in kWp.
DEFAULT_PV_UNIT: float = 1  # [kWp]

# Reference solar irradiance:
#   The reference solar irradiance, used to compute fractional PV-T electric
#   performance values.
REFERENCE_SOLAR_IRRADIANCE: float = 1000  # [W/m^2]

# Solar logger name:
#   The name to use for the solar logger.
SOLAR_LOGGER_NAME = "solar_generation"


class SolarPanelType(enum.Enum):
    """
    Specifies the type of solar panel being considered.

    - PV:
        Denotes that a PV panel is being considered.
    - PV_T:
        Denotes that a PV-T panel is being considered.

    """

    PV = "pv"
    PV_T = "pv_t"


class SolarPanel:  # pylint: disable=too-few-public-methods
    """
    Represents a solar panel being considered.

    .. attribute:: azimuthal_orientation
        The azimuthal orientation of the panel, defined in degrees from North.

    .. attribute:: lifetime
        The lifetime of the panel in years.

    .. attribute:: name
        The name of the panel being considered.

    .. attribite:: panel_type
        The type of panel being considered.

    .. attribute:: pv_unit
        The unit of PV power being considered, defaulting to 1 kWp.

    .. attribute:: pv_unit_overrided
        Whether the default PV unit was overrided (True) or not (False).

    .. attribute:: reference_efficiency
        The efficiency of the PV layer under standard test conditions.

    .. attribute:: reference_temperature
        The reference temperature of the PV layer of the panel, measured in degrees
        Celcius.

    .. attribute:: thermal_coefficient
        The thermal coefficient of performance of the PV layer of the panel, measured in
        kelvin^(-1).

    .. attribute:: tilt
        The angle between the panel and the horizontal.

    """

    panel_type: SolarPanelType

    def __init__(
        self,
        azimuthal_orientation: float,
        lifetime: int,
        name: str,
        pv_unit: float,
        pv_unit_overrided: bool,
        reference_efficiency: Optional[float],
        reference_temperature: Optional[float],
        thermal_coefficient: Optional[float],
        tilt: float,
    ) -> None:
        """
        Instantiate a :class:`SolarPanel` instance.

        Inputs:
            - azimuthal_orientation:
                The azimuthal orientation of the :class:`SolarPanel`.
            - lifetime:
                The lifetime of the :class:`SolarPanel` in years.
            - name:
                The name to assign to the :class:`SolarPanel` in order to uniquely
                identify it.
            - pv_unit:
                The output power, in Watts, of the PV layer of the panel per unit panel
                installed.
            - pv_unit_overrided:
                Whether this unit has been overrided from its default value (True) or
                not (False).
            - reference_efficiency:
                The reference efficiency of the panel, if required, otherwise `None`.
            - reference_temperature:
                The temperature, in degrees Celcius, at which the reference efficiency
                is defined, if required, otherwise `None`.
            - thermal_coefficient:
                The thermal coefficient of the PV layer of the panel, if required,
                otherwise `None`.
            - tilt:
                The tilt of the panel in degrees above the horizontal.

        """

        self.azimuthal_orientation: float = azimuthal_orientation
        self.lifetime: int = lifetime
        self.name: str = name
        self.pv_unit: float = pv_unit
        self.pv_unit_overrided: bool = pv_unit_overrided
        self.reference_efficiency: Optional[float] = reference_efficiency
        self.reference_temperature: Optional[float] = reference_temperature
        self.thermal_coefficient: Optional[float] = thermal_coefficient
        self.tilt: float = tilt

    def __init_subclass__(cls, panel_type: SolarPanelType) -> None:
        """
        The init_subclass hook, run on instantiation of the :class:`SolarPanel`.

        Inputs:
            - panel_type:
                The type of panel being considered.

        Outputs:
            An instantiated :class:`SolarPanel` instance.

        """

        cls.panel_type = panel_type

        return super().__init_subclass__()


class PVPanel(
    SolarPanel, panel_type=SolarPanelType.PV
):  # pylint: disable=too-few-public-methods
    """
    Represents a photovoltaic panel.

    """

    @classmethod
    def from_dict(cls, logger: Logger, solar_inputs: Dict[str, Any]) -> Any:
        """
        Instantiate a :class:`PVPanel` instance based on the input data.

        Inputs:
            - logger:
                The logger to use for the run.
            - solar_inputs:
                The solar input data for the panel.

        Outputs:
            A :class:`PVPanel` instance.

        """

        logger.info("Attempting to create PVPanel from solar input data.")

        if "pv_unit" in solar_inputs:
            pv_unit: float = solar_inputs["pv_unit"]
            pv_unit_overrided: bool = True
            logger.info(
                "`pv_unit` variable specified, using a pv unit of %s kWp", pv_unit
            )
        else:
            pv_unit = DEFAULT_PV_UNIT
            pv_unit_overrided = False
            logger.info("No `pv_unit` keyword specified, defaulting to %s kWp", pv_unit)

        return cls(
            solar_inputs["azimuthal_orientation"],
            solar_inputs["lifetime"],
            solar_inputs[NAME],
            pv_unit,
            pv_unit_overrided,
            solar_inputs["reference_efficiency"]
            if "reference_efficiency" in solar_inputs
            else None,
            solar_inputs["reference_temperature"]
            if "reference_temperature" in solar_inputs
            else None,
            solar_inputs["thermal_coefficient"]
            if "thermal_coefficient" in solar_inputs
            else None,
            solar_inputs["tilt"],
        )


class HybridPVTPanel(SolarPanel, panel_type=SolarPanelType.PV_T):
    """
    Represents a PV-T panel.

    .. attribute:: electric_model
        The model of the electric performance of the collector.

    .. attribute:: max_mass_flow_rate
        The maximum mass-flow rate of heat-transfer fluid through the PV-T collector,
        measured in litres per hour.

    .. attribute:: min_mass_flow_rate
        The minimum mass-flow rate of heat-transfer fluid through the PV-T collector,
        measured in litres per hour.

    .. attribute:: thermal_model
        The model of the thermal performance of the collector.

    .. attribute:: thermal_unit
        The unit of thermal panel that the panel can output which is being considered,
        measured in kWth.

    """

    def __init__(
        self,
        electric_model: Optional[Any],
        logger: Logger,
        solar_inputs: Dict[str, Any],
        solar_panels: List[SolarPanel],
        thermal_model: Optional[Any],
    ) -> None:
        """
        Instantiate a :class:`HybridPVTPanel` instance based on the input data.

        Inputs:
            - electric_model:
                The reduced electrical-efficiency model to use when generating the
                electric properties of the collector.
            - logger:
                The logger to use for the run.
            - solar_inputs:
                The solar input data specific to this panel.
            - solar_panels:
                The full set of solar generation data.
            - thermal_model:
                The reduced thermal model to use when generating the thermal properties
                of the collector.

        """

        # Attempt to extract information about the corresponding PV layer.
        try:
            pv_layer = [
                panel for panel in solar_panels if panel.name == solar_inputs["pv"]
            ][0]
        except IndexError:
            logger.error(
                "Could not find corresponding PV-layer data for layer %s for panel %s.",
                solar_inputs["pv"],
                solar_inputs[NAME],
            )
            raise InputFileError(
                "solar generation inputs",
                f"PV-layer data for layer {solar_inputs['pv']} could not be found "
                + f"whilst processing PV-T panel {solar_inputs[NAME]}.",
            ) from None

        if pv_layer.reference_efficiency is None:
            logger.error("PV reference efficiency must be defined if using PV-T.")
            raise InputFileError(
                "solar generation inputs",
                "PV reference efficiency must be defined if using PV-T",
            )
        if pv_layer.reference_temperature is None:
            logger.error("PV reference temperature must be defined if using PV-T.")
            raise InputFileError(
                "solar generation inputs",
                "PV reference temperature must be defined if using PV-T",
            )
        if pv_layer.thermal_coefficient is None:
            logger.error("PV thermal coefficient must be defined if using PV-T.")
            raise InputFileError(
                "solar generation inputs",
                "PV thermal coefficient must be defined if using PV-T",
            )

        if "pv_unit" not in solar_inputs:
            logger.error("PV unit size must be specified for PV-T panels.")
            raise InputFileError(
                "solar generation inputs",
                "PV unit size must be specified when considering PV-T panels.",
            )

        super().__init__(
            solar_inputs["azimuthal_orientation"],
            solar_inputs["lifetime"],
            solar_inputs[NAME],
            solar_inputs["pv_unit"],
            True,
            pv_layer.reference_efficiency,
            pv_layer.reference_temperature,
            pv_layer.thermal_coefficient,
            solar_inputs["tilt"],
        )

        self.electric_model = electric_model
        self.max_mass_flow_rate = solar_inputs["max_mass_flow_rate"]
        self.min_mass_flow_rate = solar_inputs["min_mass_flow_rate"]
        self.thermal_model = thermal_model
        self.thermal_unit = solar_inputs.get("thermal_unit", None)

    def __repr__(self) -> str:
        """
        Return a nice-looking representation of the panel.

        Outputs:
            - A nice-looking representation of the panel.

        """

        return (
            "HybridPVTPanel("
            + f"azimuthal_orientation={self.azimuthal_orientation}"
            + f", electric_model={self.electric_model}"
            + f", lifetime={self.lifetime}"
            + f", max_mass_flow_rate={self.max_mass_flow_rate}"
            + f", min_mass_flow_rate={self.min_mass_flow_rate}"
            + f", name={self.name}"
            + f", pv_unit={self.pv_unit}"
            + f", reference_efficiency={self.reference_efficiency}"
            + f", reference_temperature={self.reference_temperature}"
            + f", thermal_coefficient={self.thermal_coefficient}"
            + f", thermal_model={self.thermal_model}"
            + f", thermal_unit={self.thermal_unit}"
            + f", tilt={self.tilt}"
            + ")"
        )

    def calculate_performance(
        self,
        ambient_temperature: float,
        input_temperature: float,
        logger: Logger,
        mass_flow_rate: float,
        solar_irradiance: float,
        wind_speed: float,
    ) -> Tuple[float, float]:
        """
        Calculates the performance characteristics of the hybrid PV-T collector.

        The technical PV-T model developed by Benedict Winchester is reduced to a
        smaller, quick-to-run model which is loaded and utilised here.

        Inputs:
            - ambient_temperature:
                The ambient temperature, measured in degrees Celcius.
            - input_temperature:
                The input temperature of the HTF entering the PV-T collector, measured
                in degrees Celcius.
            - logger:
                The :class:`logging.Logger` to use for the run.
            - mass_flow_rate:
                The mass-flow rate of HTF passing through the collector, measured in
                kilograms per second.
            - solar_irradiance:
                The solar irradiance incident on the surface of the collector, measured
                in Watts per meter squared.
            - wind_speed:
                The wind speed at the collector, measured in meters per second.

        Outputs:
            - fractional_electric_performance:
                The fractional electric performance defined between 0 (panel is not
                operating, i.e., no output) and 1 (panel is operating at full test
                potential of reference efficiency under reference irradiance).
            - output_temperature:
                The temperature of the HTF leaving the collector, measured in degrees
                Celcius.

        """

        if self.electric_model is None or self.thermal_model is None:
            logger.error(
                "%sThe PV-T instance does not have well-defined and loaded models.%s",
                BColours.fail,
                BColours.endc,
            )
            raise ProgrammerJudgementFault(
                "pv-t modelling",
                "The PV-T instance does not have well-defined and loaded models. This "
                "could be due to the files being incorrectly parsed, mishandled, or "
                "dropped inadvertently due to internal code flow.",
            )
        if self.reference_efficiency is None:
            logger.error(
                "%sThe PV-T output function was called without a reference efficiency "
                "being defined for the PV-T panel being considered.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "solar generation inputs",
                "A reference efficiency must be defined for PV-T panels.",
            )

        input_data_frame = pd.DataFrame(
            [
                [
                    ambient_temperature,
                    input_temperature,
                    mass_flow_rate,
                    solar_irradiance,
                    wind_speed,
                ]
            ]
        )

        try:
            electric_efficiency = float(self.electric_model.predict(input_data_frame))
        except Exception as e:  # pylint: disable=broad-except
            logger.error(
                "Error attempting to predict electric efficiency of the PV-T collector: %s",
                str(e),
            )
            raise

        # Convert the efficiency to a fractional performance.
        fractional_electric_performance: float = (
            electric_efficiency / self.reference_efficiency
        ) * (solar_irradiance / REFERENCE_SOLAR_IRRADIANCE)

        try:
            output_temperature = float(self.thermal_model.predict(input_data_frame))
        except Exception as e:  # pylint: disable=broad-except
            logger.error(
                "Error attempting to predict electric efficiency of the PV-T collector: %s",
                str(e),
            )
            raise

        return fractional_electric_performance, output_temperature


def solar_degradation(lifetime: int, num_years: int) -> pd.DataFrame:
    """
    Calculates the solar degredation.

    Inputs:
        - lifetime:
            The lifetime of the solar setup in years.
        - num_years:
            The number of years for which the simulation is being carried out.

    Outputs:
        - The lifetime degredation of the solar setup.

    """

    # lifetime = self.input_data.iloc["lifetime"]
    hourly_degradation = 0.20 / (lifetime * 365 * 24)
    lifetime_degradation = []

    for i in range((num_years * 365 * 24) + 1):
        equiv = 1.0 - i * hourly_degradation
        lifetime_degradation.append(equiv)

    return pd.DataFrame(lifetime_degradation)


class SolarDataThread(
    BaseRenewablesNinjaThread, profile_name="solar", profile_key="pv"
):
    """
    Class to use when calling the solar data thread.

    """

    def __init__(
        self,
        auto_generated_files_directory: str,
        generation_inputs: Dict[str, Any],
        location: Location,
        logger_name: str,
        regenerate: bool,
        pv_panel: PVPanel,
        sleep_multiplier: int = 1,
        verbose: bool = False,
    ):
        """
        Instantiate a :class:`SolarDataThread` instance.

        """

        # Add the additional parameters which are need when calling the solar data.
        renewables_ninja_params = {
            "dataset": "merra2",
            "lat": float(location.latitude),
            "lon": float(location.longitude),
            "local_time": "false",
            "capacity": 1.0,
            "system_loss": 0,
            "tracking": 0,
            "tilt": pv_panel.tilt,
            "azim": pv_panel.azimuthal_orientation,
            "raw": "true",
        }
        super().__init__(
            auto_generated_files_directory,
            generation_inputs,
            location,
            logger_name,
            regenerate,
            sleep_multiplier,
            verbose,
            renewables_ninja_params=renewables_ninja_params,
        )


def total_solar_output(*args, **kwargs) -> pd.DataFrame:  # type: ignore
    """
    Wrapper function to wrap the total solar output.

    """

    return total_profile_output(*args, **kwargs, profile_name="solar")



#!/usr/bin/python3
########################################################################################
# appraisal.py - Optimisation appraisal module.                                        #
#                                                                                      #
# Authors: Phil Sandwell, Ben Winchester                                               #
# Copyright: Phil Sandwell, 2018                                                       #
# License: Open source                                                                 #
# Most recent update: 05/08/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
appraisal.py - The optimisation appraisal module of CLOVER.

In order to compare and evaluate energy systems, appraisals of their performance need to
be carried out. This module exposes methods to appraise the results of CLOVER
simulations.

"""

from logging import Logger
from typing import Any, Dict, Optional

import numpy as np  # pylint: disable=import-error
import pandas as pd

from ..impact import finance, ghgs

from ..__utils__ import (
    BColours,
    ColumnHeader,
    Criterion,
    CumulativeResults,
    EnvironmentalAppraisal,
    FinancialAppraisal,
    InternalError,
    hourly_profile_to_daily_sum,
    Location,
    Scenario,
    SystemAppraisal,
    SystemDetails,
    TechnicalAppraisal,
)
from ..impact.__utils__ import ImpactingComponent

__all__ = ("appraise_system",)


def _simulation_environmental_appraisal(  # pylint: disable=too-many-locals
    buffer_tank_addition: int,
    clean_water_tank_addition: int,
    converter_addition: Dict[str, int],
    diesel_addition: float,
    electric_yearly_load_statistics: pd.DataFrame,
    end_year: int,
    ghg_inputs: Dict[str, Any],
    heat_exchanger_addition: int,
    hot_water_tank_addition: int,
    location: Location,
    logger: Logger,
    pv_addition: float,
    pvt_addition: float,
    scenario: Scenario,
    simulation_results: pd.DataFrame,
    start_year: int,
    storage_addition: float,
    system_details: SystemDetails,
) -> EnvironmentalAppraisal:
    """
    Appraises the environmental impact of a minigrid system

    Inputs:
        - buffer_tank_addition:
            The additional number of buffer tanks added this iteration.
        - clean_water_tank_addition:
            The additional number of clean-water tanks added this iteration.
        - converter_addition:
            A mapping between converter names and the size of each that was added to the
            system this iteration.
        - diesel_addition:
            The additional diesel capacity added this iteration.
        - electric_yearly_load_statistics:
            The yearly electric load statistics.
        - end_year:
            The end year of the simulation period.
        - ghg_inputs:
            The GHG input information.
        - heat_exchanger_addition:
            The additional number of heat exchangers added this iteration.
        - hot_water_tank_addition:
            The additional number of hot-water tanks added this iteration.
        - location:
            The location being considered.
        - logger:
            The logger to use for the run.
        - pv_addition:
            The additional number of PV panels added this iteration.
        - pvt_addition:
            The additional number of PV-T panels added this iteration.
        - scenario:
            The :class:`Scenario` currently being considered.
        - simulation_results:
            The system that was just simulated.
        - start_year:
            The start year for the simulation period.
        - storage_addition:
            The additional number of electric storage devices (batteries) added this
            iteration.
        - system_details:
            The deatils of the system that was just simulated.

    Outputs:
        An :class:`EnvironmentalAppraisal` containing the key environmental results.

    """

    # Calculate new equipment GHGs
    try:
        equipment_ghgs = ghgs.calculate_total_equipment_ghgs(
            buffer_tank_addition,
            clean_water_tank_addition,
            converter_addition,
            diesel_addition,
            ghg_inputs,
            heat_exchanger_addition,
            hot_water_tank_addition,
            logger,
            pv_addition,
            pvt_addition,
            storage_addition,
        ) + ghgs.calculate_independent_ghgs(
            electric_yearly_load_statistics,
            end_year,
            ghg_inputs,
            location,
            logger,
            scenario,
            start_year,
        )
    except KeyError as e:
        logger.error("Missing system equipment GHG input information: %s", str(e))
        raise

    # Calculate GHGs of connecting new households
    try:
        connections_ghgs = ghgs.calculate_connections_ghgs(
            ghg_inputs, simulation_results[ColumnHeader.HOUSEHOLDS.value]
        )
    except KeyError as e:
        logger.error("Missing household connection GHG input information: %s", str(e))
        raise

    # Calculate operating GHGs of the system during this simulation
    try:
        om_ghgs = ghgs.calculate_total_om(
            system_details.initial_num_buffer_tanks
            if system_details.initial_num_buffer_tanks is not None
            else 0,
            system_details.initial_num_clean_water_tanks
            if system_details.initial_num_clean_water_tanks is not None
            else 0,
            system_details.initial_converter_sizes
            if system_details.initial_converter_sizes is not None
            else None,
            system_details.diesel_capacity,
            ghg_inputs,
            system_details.initial_num_buffer_tanks
            if system_details.initial_num_buffer_tanks is not None
            else 0,
            system_details.initial_num_hot_water_tanks
            if system_details.initial_num_hot_water_tanks is not None
            else 0,
            logger,
            system_details.initial_pv_size,
            system_details.initial_pvt_size
            if system_details.initial_pvt_size is not None
            else 0,
            system_details.initial_storage_size,
            start_year,
            end_year,
        )
    except KeyError as e:
        logger.error("Missing O&M GHG input information: %s", str(e))
        raise

    # Calculate running GHGs of the system
    try:
        diesel_fuel_ghgs = ghgs.calculate_diesel_fuel_ghgs(
            simulation_results[ColumnHeader.DIESEL_FUEL_USAGE.value], ghg_inputs
        )
    except KeyError as e:
        logger.error("Missing diesel-fuel GHG input information: %s", str(e))
        raise

    try:
        grid_ghgs = ghgs.calculate_grid_ghgs(
            ghg_inputs,
            simulation_results[ColumnHeader.GRID_ENERGY.value],
            location,
            start_year,
            end_year,
        )
    except KeyError as e:
        logger.error("Missing grid GHG input information: %s", str(e))
        raise
    try:
        kerosene_ghgs = ghgs.calculate_kerosene_ghgs(
            ghg_inputs, simulation_results[ColumnHeader.KEROSENE_LAMPS.value]
        )
    except KeyError as e:
        logger.error("Missing kerosene GHG input information: %s", str(e))
        raise

    try:
        kerosene_ghgs_mitigated = ghgs.calculate_kerosene_ghgs_mitigated(
            ghg_inputs, simulation_results[ColumnHeader.KEROSENE_MITIGATION.value]
        )
    except KeyError as e:
        logger.error("Missing kerosene GHG input information: %s", str(e))
        raise

    # Total GHGs incurred during simulation period
    total_ghgs = (
        equipment_ghgs
        + connections_ghgs
        + om_ghgs
        + diesel_fuel_ghgs
        + grid_ghgs
        + kerosene_ghgs
    )
    total_system_ghgs = (
        equipment_ghgs + connections_ghgs + om_ghgs + diesel_fuel_ghgs + grid_ghgs
    )

    # Return outputs
    return EnvironmentalAppraisal(
        round(diesel_fuel_ghgs, 3),
        round(grid_ghgs, 3),
        round(kerosene_ghgs, 3),
        round(kerosene_ghgs_mitigated, 3),
        round(connections_ghgs, 3),
        round(equipment_ghgs, 3),
        round(om_ghgs, 3),
        round(total_ghgs, 3),
        round(total_system_ghgs, 3),
    )


def _simulation_financial_appraisal(  # pylint: disable=too-many-locals
    buffer_tank_addition: int,
    clean_water_tank_addition: int,
    converter_addition: Dict[str, int],
    diesel_addition: float,
    finance_inputs: Dict[str, Any],
    heat_exchanger_addition: int,
    hot_water_tank_addition: int,
    location: Location,
    logger: Logger,
    pv_addition: float,
    pvt_addition: float,
    scenario: Scenario,
    simulation_results: pd.DataFrame,
    storage_addition: float,
    system_details: SystemDetails,
    yearly_load_statistics: pd.DataFrame,
) -> FinancialAppraisal:
    """
    Appraises the financial performance of a minigrid system.

    Inputs:
        - buffer_tank_addition:
            The additional number of buffer tanks added this iteration.
        - clean_water_tank_addition:
            The additional number of clean-water tanks added this iteration.
        - converter_addition:
            A mapping between converter names and the size of each that was added to the
            system this iteration.
        - diesel_addition:
            The additional diesel capacity added this iteration.
        - finance_inputs:
            The finance input information.
        - heat_exchanger_addition:
            The additional number of heat exchangers added this iteration.
        - hot_water_tank_addition:
            The additional number of hot-water tanks added this iteration.
        - location:
            The :class:`Location` being considered.
        - logger:
            The logger to use for the run.
        - pv_addition:
            The additional number of PV panels added this iteration.
        - pvt_addition:
            The additional number of PV-T panels added this iteration.
        - scenario:
            The :class:`Scenario` currently being considered.
        - simulation_results:
            Outputs of Energy_System().simulation(...)
        - storage_addition:
            The additional number of electric storage devices (batteries) added this
            iteration.
        - system_details:
            The details of this system.
        - yearly_load_statistics:
            The yearly electric load statistics for the system.

    Outputs:
        The financial appraisal of the system.

    """

    # Calculate new equipment costs (discounted)
    equipment_costs = finance.discounted_equipment_cost(
        buffer_tank_addition,
        clean_water_tank_addition,
        converter_addition,
        diesel_addition,
        finance_inputs,
        heat_exchanger_addition,
        hot_water_tank_addition,
        logger,
        pv_addition,
        pvt_addition,
        storage_addition,
        system_details.start_year,
    ) + finance.independent_expenditure(
        finance_inputs,
        location,
        logger,
        scenario,
        yearly_load_statistics,
        start_year=system_details.start_year,
        end_year=system_details.end_year,
    )

    # Calculate costs of connecting new households (discounted)
    connections_cost = finance.connections_expenditure(
        finance_inputs,
        simulation_results[ColumnHeader.HOUSEHOLDS.value],
        system_details.start_year,
    )

    # Calculate operating costs of the system during this simulation (discounted)
    om_costs = finance.total_om(
        system_details.initial_num_buffer_tanks
        if system_details.initial_num_buffer_tanks is not None
        else 0,
        system_details.initial_num_clean_water_tanks
        if system_details.initial_num_clean_water_tanks is not None
        else 0,
        system_details.initial_converter_sizes
        if system_details.initial_converter_sizes is not None
        else None,
        system_details.diesel_capacity,
        finance_inputs,
        system_details.initial_num_buffer_tanks
        if system_details.initial_num_buffer_tanks is not None
        else 0,
        system_details.initial_num_hot_water_tanks
        if system_details.initial_num_hot_water_tanks is not None
        else 0,
        logger,
        system_details.initial_pv_size,
        system_details.initial_pvt_size
        if system_details.initial_pvt_size is not None
        else 0,
        system_details.initial_storage_size,
        start_year=system_details.start_year,
        end_year=system_details.end_year,
    )

    # Calculate running costs of the system (discounted)
    diesel_costs = finance.diesel_fuel_expenditure(
        simulation_results[ColumnHeader.DIESEL_FUEL_USAGE.value],
        finance_inputs,
        logger,
        start_year=system_details.start_year,
        end_year=system_details.end_year,
    )
    grid_costs = finance.expenditure(
        ImpactingComponent.GRID,
        finance_inputs,
        simulation_results[ColumnHeader.GRID_ENERGY.value],
        logger,
        start_year=system_details.start_year,
        end_year=system_details.end_year,
    )
    kerosene_costs = finance.expenditure(
        ImpactingComponent.KEROSENE,
        finance_inputs,
        simulation_results[ColumnHeader.KEROSENE_LAMPS.value],
        logger,
        start_year=system_details.start_year,
        end_year=system_details.end_year,
    )
    kerosene_costs_mitigated = finance.expenditure(
        ImpactingComponent.KEROSENE,
        finance_inputs,
        simulation_results[ColumnHeader.KEROSENE_MITIGATION.value],
        logger,
        start_year=system_details.start_year,
        end_year=system_details.end_year,
    )

    # Total cost incurred during simulation period (discounted)
    total_cost = (
        equipment_costs
        + connections_cost
        + om_costs
        + diesel_costs
        + grid_costs
        + kerosene_costs
    )
    total_system_cost = (
        equipment_costs + connections_cost + om_costs + diesel_costs + grid_costs
    )

    # Return outputs
    return FinancialAppraisal(
        round(diesel_costs, 3),
        round(grid_costs, 3),
        round(kerosene_costs, 3),
        round(kerosene_costs_mitigated, 3),
        round(connections_cost, 3),
        round(equipment_costs, 3),
        round(om_costs, 3),
        round(total_cost, 3),
        round(total_system_cost, 3),
    )


def _simulation_technical_appraisal(  # pylint: disable=too-many-locals
    finance_inputs: Dict[str, Any],
    logger: Logger,
    simulation_results: pd.DataFrame,
    system_details: SystemDetails,
) -> TechnicalAppraisal:
    """
    Appraises the technical performance of a minigrid system

    Inputs:
        - simulation:
            Outputs of Energy_System().simulation(...).
        - start_year:
            Start year of this simulation period.
        - end_year:
            End year of this simulation period.

    Outputs:
        - system_outputs:
            A :class:`pd.DataFrame` containing key technical data e.g. energy used,
            unmet energy, blackout percentage, discounted energy.

    """

    # Calculate system blackouts
    system_blackouts: float = float(
        np.mean(simulation_results[ColumnHeader.BLACKOUTS.value].values)  # type: ignore
    )

    # Clean-water system.
    clean_water_blackouts: Optional[float] = (
        round(
            float(
                np.mean(
                    simulation_results[  # type: ignore
                        ColumnHeader.CLEAN_WATER_BLACKOUTS.value
                    ].values
                )
            ),
            3,
        )
        if ColumnHeader.CLEAN_WATER_BLACKOUTS.value in simulation_results
        else None
    )
    total_clean_water = (
        np.sum(simulation_results[ColumnHeader.TOTAL_CW_SUPPLIED.value])  # type: ignore
        if ColumnHeader.TOTAL_CW_SUPPLIED.value in simulation_results
        else 0
    )

    # Energy system.
    total_energy = np.sum(
        simulation_results[ColumnHeader.TOTAL_ELECTRICITY_CONSUMED.value]  # type: ignore
    )
    total_load_energy = np.sum(
        simulation_results[ColumnHeader.LOAD_ENERGY.value]  # type: ignore
    )
    total_renewables_used = np.sum(
        simulation_results[ColumnHeader.RENEWABLE_ELECTRICITY_USED_DIRECTLY.value]  # type: ignore
    )
    total_pv_energy = np.sum(
        simulation_results[ColumnHeader.PV_ELECTRICITY_SUPPLIED.value]  # type: ignore
    )
    total_pvt_energy = (
        np.sum(
            simulation_results[ColumnHeader.TOTAL_PVT_ELECTRICITY_SUPPLIED.value]  # type: ignore
        )
        if ColumnHeader.TOTAL_PVT_ELECTRICITY_SUPPLIED.value in simulation_results
        else None
    )
    total_storage_used = np.sum(
        simulation_results[ColumnHeader.ELECTRICITY_FROM_STORAGE.value]  # type: ignore
    )
    total_grid_used = np.sum(
        simulation_results[ColumnHeader.GRID_ENERGY.value]  # type: ignore
    )
    total_diesel_used = np.sum(
        simulation_results[ColumnHeader.DIESEL_ENERGY_SUPPLIED.value]  # type: ignore
    )
    total_unmet_energy = np.sum(
        simulation_results[ColumnHeader.UNMET_ELECTRICITY.value]  # type: ignore
    )
    renewables_fraction = (total_renewables_used + total_storage_used) / total_energy
    unmet_fraction = total_unmet_energy / total_load_energy

    # Calculate total discounted energy
    total_energy_daily = hourly_profile_to_daily_sum(
        pd.DataFrame(simulation_results[ColumnHeader.TOTAL_ELECTRICITY_CONSUMED.value])
    )
    discounted_energy = finance.discounted_energy_total(
        finance_inputs,
        logger,
        total_energy_daily,
        start_year=system_details.start_year,
        end_year=system_details.end_year,
    )

    # Calculate proportion of kerosene displaced (defaults to zero if kerosene is not
    # originally used
    if np.sum(simulation_results[ColumnHeader.KEROSENE_LAMPS.value]) > 0.0:  # type: ignore
        kerosene_displacement = (
            np.sum(
                simulation_results[ColumnHeader.KEROSENE_MITIGATION.value]  # type: ignore
            )
        ) / (
            np.sum(
                simulation_results[ColumnHeader.KEROSENE_MITIGATION.value]  # type: ignore
            )
            + np.sum(
                simulation_results[ColumnHeader.KEROSENE_LAMPS.value]  # type: ignore
            )
        )
    else:
        kerosene_displacement = 0.0

    # Calculate diesel fuel usage
    total_diesel_fuel = np.sum(
        simulation_results[ColumnHeader.DIESEL_FUEL_USAGE.value]  # type: ignore
    )

    # Return outputs
    return TechnicalAppraisal(
        round(system_blackouts, 3),
        clean_water_blackouts,
        round(total_diesel_used, 3),
        round(total_diesel_fuel, 3),
        round(discounted_energy, 3),
        round(total_grid_used, 3),
        round(kerosene_displacement, 3),
        round(total_pv_energy, 3),
        round(total_pvt_energy, 3) if total_pvt_energy is not None else None,
        round(total_renewables_used, 3),
        round(renewables_fraction, 3),
        round(total_storage_used, 3),
        round(total_clean_water, 3),
        round(total_energy, 3),
        round(total_unmet_energy, 3),
        round(unmet_fraction, 3),
    )


def appraise_system(  # pylint: disable=too-many-locals
    electric_yearly_load_statistics: pd.DataFrame,
    end_year: int,
    finance_inputs: Dict[str, Any],
    ghg_inputs: Dict[str, Any],
    location: Location,
    logger: Logger,
    previous_system: Optional[SystemAppraisal],
    scenario: Scenario,
    simulation_results: pd.DataFrame,
    start_year: int,
    system_details: SystemDetails,
) -> SystemAppraisal:
    """
    Appraises the total performance of a minigrid system for all performance metrics

    Inputs:
        - electric_yearly_load_statistics:
            The yearly electric load statistics.
        - end_year:
            The end year for the simulation that was just run.
        - finance_inputs:
            The finance input information.
        - location:
            The location currently being considered.
        - logger:
            The logger to use for the run.
        - previous_system:
            Report from previously installed system (not required if no system was
            previously deployed)
        - scenario:
            The :class:`Scenario` currently being considered.
        - simulation_results
            Outputs of Energy_System().simulation(...)
        - start_year:
            The start year for the simulation that was just run.
        - system_details:
            The system details about the system that was just simulated.

    Outputs:
        - system_outputs:
            :class:`pd.DataFrame` containing all key technical, performance,
            financial and environmental information.

    """

    if previous_system is None:
        previous_system = SystemAppraisal(
            CumulativeResults(),
            EnvironmentalAppraisal(),
            FinancialAppraisal(),
            SystemDetails(),
            TechnicalAppraisal(),
        )

    # Compute the additions made to the system.
    buffer_tank_addition: int = (
        system_details.initial_num_buffer_tanks
        - previous_system.system_details.final_num_buffer_tanks
        if system_details.initial_num_buffer_tanks is not None
        and previous_system.system_details.final_num_buffer_tanks is not None
        else 0
    )
    clean_water_tank_addition: int = (
        system_details.initial_num_clean_water_tanks
        - previous_system.system_details.final_num_clean_water_tanks
        if system_details.initial_num_clean_water_tanks is not None
        and previous_system.system_details.final_num_clean_water_tanks is not None
        else 0
    )
    if system_details.initial_converter_sizes is None:
        logger.error(
            "%sNo converter sizes on system details when calling system appraisal. "
            "Only systems that have been simulated can be appraised.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InternalError("Misuse of system appraisal function.")
    converter_addition: Dict[str, int] = {
        converter: size
        - (
            previous_system.system_details.final_converter_sizes[converter]
            if previous_system.system_details.final_converter_sizes is not None
            else 0
        )
        for converter, size in system_details.initial_converter_sizes.items()
    }
    diesel_addition = (
        system_details.diesel_capacity - previous_system.system_details.diesel_capacity
    )
    heat_exchanger_addition: int = (
        system_details.initial_num_buffer_tanks
        - previous_system.system_details.final_num_buffer_tanks
        if system_details.initial_num_buffer_tanks is not None
        and previous_system.system_details.final_num_buffer_tanks is not None
        else 0
    )
    hot_water_tank_addition: int = (
        system_details.initial_num_hot_water_tanks
        - previous_system.system_details.final_num_hot_water_tanks
        if system_details.initial_num_hot_water_tanks is not None
        and previous_system.system_details.final_num_hot_water_tanks is not None
        else 0
    )
    pv_addition = (
        system_details.initial_pv_size - previous_system.system_details.final_pv_size
    )
    pvt_addition: float = (
        system_details.initial_pvt_size - previous_system.system_details.final_pvt_size
        if system_details.initial_pvt_size is not None
        and previous_system.system_details.final_pvt_size is not None
        else 0
    )
    storage_addition = (
        system_details.initial_storage_size
        - previous_system.system_details.final_storage_size
    )

    # Get results which will be carried forward into optimisation process
    technical_appraisal = _simulation_technical_appraisal(
        finance_inputs, logger, simulation_results, system_details
    )

    financial_appraisal = _simulation_financial_appraisal(
        buffer_tank_addition,
        clean_water_tank_addition,
        converter_addition,
        diesel_addition,
        finance_inputs,
        heat_exchanger_addition,
        hot_water_tank_addition,
        location,
        logger,
        pv_addition,
        pvt_addition,
        scenario,
        simulation_results,
        storage_addition,
        system_details,
        electric_yearly_load_statistics,
    )
    environmental_appraisal = _simulation_environmental_appraisal(
        buffer_tank_addition,
        clean_water_tank_addition,
        converter_addition,
        diesel_addition,
        electric_yearly_load_statistics,
        end_year,
        ghg_inputs,
        heat_exchanger_addition,
        hot_water_tank_addition,
        location,
        logger,
        pv_addition,
        pvt_addition,
        scenario,
        simulation_results,
        start_year,
        storage_addition,
        system_details,
    )

    # Get results that rely on metrics of different kinds and several different iteration periods
    if (
        technical_appraisal.total_clean_water > 0
        and previous_system.cumulative_results.clean_water is not None
    ):
        cumulative_clean_water: float = (
            technical_appraisal.total_clean_water
            + previous_system.cumulative_results.clean_water
        )
    else:
        logger.debug("No clean water produced.")
        cumulative_clean_water = 0
    cumulative_costs = (
        financial_appraisal.total_cost + previous_system.cumulative_results.cost
    )
    cumulative_discounted_energy = (
        technical_appraisal.discounted_energy
        + previous_system.cumulative_results.discounted_energy
    )
    cumulative_energy = (
        technical_appraisal.total_energy + previous_system.cumulative_results.energy
    )
    cumulative_ghgs = (
        environmental_appraisal.total_ghgs + previous_system.cumulative_results.ghgs
    )
    cumulative_system_costs = (
        financial_appraisal.total_system_cost
        + previous_system.cumulative_results.system_cost
    )
    cumulative_system_ghgs = (
        environmental_appraisal.total_system_ghgs
        + previous_system.cumulative_results.system_ghgs
    )

    # Combined metrics
    lcue = float(cumulative_system_costs / cumulative_discounted_energy)
    # lcuw = float(cumulative_system_costs / cumulative_discounted_clean_water)
    emissions_intensity = 1000.0 * float(cumulative_system_ghgs / cumulative_energy)

    #   Format outputs
    cumulative_results = CumulativeResults(
        cumulative_clean_water,
        cumulative_costs,
        cumulative_discounted_energy,
        cumulative_energy,
        cumulative_ghgs,
        cumulative_system_costs,
        cumulative_system_ghgs,
    )

    criteria = {
        Criterion.BLACKOUTS: round(technical_appraisal.blackouts, 3),
        Criterion.CUMULATIVE_COST: round(cumulative_results.cost, 3),
        Criterion.CUMULATIVE_GHGS: round(cumulative_results.ghgs, 3),
        Criterion.CUMULATIVE_SYSTEM_COST: round(cumulative_results.system_cost, 3),
        Criterion.CUMULATIVE_SYSTEM_GHGS: round(cumulative_results.system_ghgs, 3),
        Criterion.EMISSIONS_INTENSITY: round(emissions_intensity, 3),
        Criterion.KEROSENE_COST_MITIGATED: round(
            financial_appraisal.kerosene_cost_mitigated, 3
        ),
        Criterion.KEROSENE_DISPLACEMENT: round(
            technical_appraisal.kerosene_displacement, 3
        ),
        Criterion.KEROSENE_GHGS_MITIGATED: round(
            environmental_appraisal.kerosene_ghgs_mitigated, 3
        ),
        Criterion.LCUE: round(lcue, 3),
        Criterion.RENEWABLES_FRACTION: round(
            technical_appraisal.renewable_energy_fraction, 3
        ),
        Criterion.TOTAL_COST: round(financial_appraisal.total_cost, 3),
        Criterion.TOTAL_GHGS: round(environmental_appraisal.total_ghgs, 3),
        Criterion.TOTAL_SYSTEM_COST: round(financial_appraisal.total_system_cost, 3),
        Criterion.TOTAL_SYSTEM_GHGS: round(
            environmental_appraisal.total_system_ghgs, 3
        ),
        Criterion.UNMET_ENERGY_FRACTION: round(
            technical_appraisal.unmet_energy_fraction, 3
        ),
    }

    if technical_appraisal.clean_water_blackouts is not None:
        criteria[Criterion.CLEAN_WATER_BLACKOUTS] = round(
            technical_appraisal.clean_water_blackouts, 3
        )

    # Combine the outputs into a single system appraisal instance.
    system_appraisal = SystemAppraisal(
        cumulative_results,
        environmental_appraisal,
        financial_appraisal,
        system_details,
        technical_appraisal,
        criteria=criteria,
    )

    return system_appraisal



#!/usr/bin/python3
########################################################################################
# __utils__.py - CLOVER Optimisation Utility module.                                   #
#                                                                                      #
# Author: Phil Sandwell, Ben Winchester                                                #
# Copyright: Phil Sandwell, 2018                                                       #
# Date created: 24/08/2021                                                             #
#                                                                                      #
# For more information, please email:                                                  #
#   philip.sandwell@gmail.com                                                          #
########################################################################################
"""
__utils__.py - Utility module for CLOVER's Optimisation Component.

The utility module contains functionality which is used by the optimisation and
appraisal modules within the optimisation component.

"""

import dataclasses
import enum
import os

from logging import Logger
from typing import Any, Dict, List, Optional, Pattern, Tuple, Union

import json
import re

import pandas as pd  # pylint: disable=import-error
from tqdm import tqdm

from ..simulation import energy_system

from ..__utils__ import (
    DEFAULT_SCENARIO,
    BColours,
    Criterion,
    InputFileError,
    ITERATION_LENGTH,
    Location,
    MAX,
    MIN,
    NUMBER_OF_ITERATIONS,
    ProgrammerJudgementFault,
    RenewableEnergySource,
    ResourceType,
    Scenario,
    Simulation,
    STEP,
    SystemAppraisal,
)
from ..conversion.conversion import Converter, WaterSource
from ..impact.__utils__ import ImpactingComponent

from .appraisal import appraise_system

__all__ = (
    "converters_from_sizing",
    "ConverterSize",
    "CriterionMode",
    "get_sufficient_appraisals",
    "Optimisation",
    "OptimisationParameters",
    "recursive_iteration",
    "save_optimisation",
    "SolarSystemSize",
    "StorageSystemSize",
    "TankSize",
    "THRESHOLD_CRITERIA",
    "ThresholdMode",
)

# Converter name string:
#   The name used for parsing the converter name group.
#   NOTE: This name is not updated within the regex and needs to be updated separately.
CONVERTER_NAME_STRING: str = "name"

# Converter size regex:
#   Regular expression used for parsing the size of various converters for
# optimisations.
#   NOTE: The name of the group is not updated automatically in accordance with the
# above string and needs to be udpated separately.
CONVERTER_SIZE_REGEX: Pattern[str] = re.compile(r"(?P<name>.*)_size")

# Optimisation criteria:
#   Keyword used for parsing the optimisation criteria.
OPTIMISATION_CRITERIA: str = "optimisation_criteria"

# Scenario:
#   Keyword used for parsing the scenario to use for a given optimisation.
SCENARIO: str = "scenario"

# Threshold criteria:
#   Keyword used for parsing the threshold criteria.
THRESHOLD_CRITERIA: str = "threshold_criteria"


def converters_from_sizing(converter_sizes: Dict[Converter, int]) -> List[Converter]:
    """
    Generates a `list` of available converters based on the number of each available.

    As the system is optimised, it becomes necessary to generate a `list` containing the
    available converters, with duplicates allowed to indiciate multiple instances of a
    single type present, from the various values.

    Inputs:
        - converter_sizes:
            A `dict` mapping :class:`Converter` instances to the number of each type
            present during the iteration.

    Outputs:
        - A `list` of :class:`Converter` instances present based on the mapping passed
        in.

    """

    converters: List[Converter] = []

    for converter, size in converter_sizes.items():
        converters.extend([converter] * size)

    return converters


@dataclasses.dataclass
class ConverterSize:
    """
    Used to wrap the converter size information.

    .. atttribute:: max
        The maximum size of the :class:`converseion.Converter` in question, measured in
        number of :class:`conversion.Converter` instances.

    .. attribute:: min
        The minimum size of the :class:`converseion.Converter` in question, measured in
        number of :class:`conversion.Converter` instances.

    .. attribute:: step
        The step to use for the :class:`converseion.Converter` in question, measured in
        number of :class:`conversion.Converter` instances.

    """

    max: int = 0
    min: int = 0
    step: int = 1


class CriterionMode(enum.Enum):
    """
    The mode of optimisation of the criterion.

    - MAXIMISE:
        Denotes that the optimisation criterion is to be maximised.

    - MINIMISE:
        Denotes that the optimisation criterion is to be minimised.

    """

    MAXIMISE = "maximise"
    MINIMISE = "minimise"

    def __str__(self) -> str:
        """
        Returns a nice-looking `str` representing the :class:`CriterionMode`.

        Outputs:
            - A nice-looking `str` representing the :class:`CriterionMode`
              instance.

        """

        return f"CriterionMode(mode={self.value})"


class ThresholdMode(enum.Enum):
    """
    Represents whether a threshold value is a maximum or minimum limit on the system.

    - MAXIMUM:
        Denotes that the threshold is a maximum to be placed on the system.
    - MINIMUM:
        Denotes that the threshold is a minimum to be placed on the system.

    """

    MAXIMUM = "maximum"
    MINIMUM = "minimum"


@dataclasses.dataclass
class Optimisation:
    """
    Represents an optimisation to be carried out.

    .. attribute:: optimisation_criteria
        A `dict` mapping optimisation criteria to whether they should be maximised or
        minimised.

    .. attribute:: scenario
        The :class:`Scenario` to use for this optimisation.

    .. attribute:: threshold_criteria
        A `dict` mapping threshold criteria to their values.

    """

    optimisation_criteria: Dict[Criterion, CriterionMode]
    scenario: Scenario
    threshold_criteria: Dict[Criterion, float]

    def __str__(self) -> str:
        """
        Returns a nice-looking `str` summarising the :class:`Optimisation` instance.

        Outputs:
            - A nice-looking `str` summarising the information contained within the
              :class:`Optimisation` instance.

        """

        return (
            "Optimisation("
            + f"optimisation_crtieria: {self.optimisation_criteria}"
            + f", scenario: {self.scenario}"
            + f", threshold_criteria: {self.threshold_criteria}"
            + ")"
        )

    def __hash__(self) -> int:
        """
        Returns an `int` representing the :class:`Optimisation` instance for sorting.

        In order to efficiently store :class:`Optimisation` instances, a unique hash is
        required. This method computes such a hash and returns it.

        Outputs:
            - A unique `int` representing the :class:`Optimisation` instance.

        """

        return hash(str(self))

    @classmethod
    def from_dict(
        cls,
        logger: Logger,
        optimisation_data: Dict[str, Any],
        scenarios: List[Scenario],
    ) -> Any:
        """
        Creates a :class:`Optimisation` instance based on the input data.

        Inputs:
            - logger:
                The logger to use for the run.
            - optimisation_data:
                The optimisation data, extracted from the input file.
            - scenarios:
                The `list` of :class:`Scenario` instances available for the run.

        Outputs:
            - A :class:`Optimisation` instance based on the input data.

        """

        try:
            optimisation_criteria = {
                Criterion(key): CriterionMode(value)
                for entry in optimisation_data[OPTIMISATION_CRITERIA]
                for key, value in entry.items()
            }
        except KeyError as e:
            logger.error(
                "%sError processing optimisation criteria, missing entry: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise

        try:
            threshold_criteria = {
                Criterion(key): value
                for entry in optimisation_data[THRESHOLD_CRITERIA]
                for key, value in entry.items()
            }
        except KeyError as e:
            logger.error(
                "%sError processing threshold criteria, missing entry: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise

        if SCENARIO in optimisation_data:
            try:
                scenario = [
                    scenario
                    for scenario in scenarios
                    if scenario.name == optimisation_data["scenario"]
                ][0]
            except IndexError:
                logger.error(
                    "%sError determining scenario for optimisation run: scenario '%s' "
                    "could not be found.%s",
                    BColours.fail,
                    optimisation_data["scenario"],
                    BColours.endc,
                )
                raise InputFileError(
                    "optimisation inputs/scenario inputs",
                    f"Scenario {optimisation_data['scenario']} could not be found.",
                ) from None
        else:
            try:
                scenario = [
                    scenario
                    for scenario in scenarios
                    if scenario.name == DEFAULT_SCENARIO
                ][0]
            except IndexError:
                logger.error(
                    "%sError determining scenario for optimisation run: default "
                    "scenario '%s' could not be found.%s",
                    BColours.fail,
                    DEFAULT_SCENARIO,
                    BColours.endc,
                )
                raise InputFileError(
                    "optimisation inputs/scenario inputs",
                    f"Default scenario {DEFAULT_SCENARIO} could not be found.",
                ) from None

        return cls(optimisation_criteria, scenario, threshold_criteria)

    def to_dict(self) -> Dict[str, Any]:
        """
        Returns a `dict` summarising the :class:`Optimisation` instance.

        Outputs:
            - A `dict` summarising the information contained within the
              :class:`Optimisation` instance.

        """

        optimisation_criteria = {
            str(key.value): str(value.value)
            for key, value in self.optimisation_criteria.items()
        }
        threshold_criteria = {
            str(key.value): float(value)
            for key, value in self.threshold_criteria.items()
        }

        return {
            OPTIMISATION_CRITERIA: optimisation_criteria,
            "scenario": self.scenario.to_dict(),
            THRESHOLD_CRITERIA: threshold_criteria,
        }


@dataclasses.dataclass
class OptimisationComponent(enum.Enum):
    """
    Contains information about the components which are variable in an optimisation.

    - CLEAN_WATER_PVT_SIZE:
        Denotes the size of the clean-water PV-T system, measured in PV-T units.

    - CLEAN_WATER_TANKS:
        Denotes the number of clean-water tanks in the system.

    - HOT_WATER_PVT_SIZE:
        Denotes the size of the hot-water PV-T system, measured in PV-T units.

    - HOT_WATER_TANKS:
        Denotes the number of hot-water tanks in the system.

    - PV_SIZE:
        Denotes the size of the PV system, measured in PV units.

    - STORAGE_SIZE:
        Denotes the size of the storage system, measured in storage units, i.e.,
        batteries.

    """

    CLEAN_WATER_PVT_SIZE = "cw_pvt_size"
    CLEAN_WATER_TANKS = "cw_tanks"
    HOT_WATER_PVT_SIZE = "hw_pvt_size"
    HOT_WATER_TANKS = "hw_tanks"
    PV_SIZE = "pv_size"
    STORAGE_SIZE = "storage_size"


@dataclasses.dataclass
class SolarSystemSize:
    """
    Used to wrap the solar-system-size information.

    .. attribute:: max
        The maximum size of the system, measured in PV or PV-T units.

    .. attribute:: min
        The minimum size of the system, measured in PV or PV-T units.

    .. attribute:: step
        The step to use for the system, measured in PV or PV-T units.

    """

    max: float = 0
    min: float = 0
    step: float = 1


@dataclasses.dataclass
class StorageSystemSize:
    """
    Used to wrap the storage-system-size information.

    .. attribute:: max
        The maximum size of the system, measured in storage units, i.e., batteries.

    .. attribute:: min
        The minimum size of the system, measured in storage units, i.e., batteries.

    .. attribute:: step
        The step to use for the system, measured in storage units, i.e., batteries.

    """

    max: float = 0
    min: float = 0
    step: float = 1


@dataclasses.dataclass
class TankSize:
    """
    Used to wrap the tank size information.

    .. atttribute:: max
        The maximum size of the tank system, measured in number of tanks.

    .. attribute:: min
        The minimum size of the tank system, measured in number of tanks.

    .. attribute:: step
        The step to use for the tank system, measured in number of tanks.

    """

    max: int = 0
    min: int = 0
    step: int = 1


@dataclasses.dataclass
class OptimisationParameters:
    """
    Parameters that define the scope of the optimisation.

    .. attribute:: clean_water_tanks
        The sizing bounds for the clean-water tanks.

    .. attribute:: converter_sizes
        The sizing bounds for the various :class:`conversion.Converter` instances
        associated with the system.

    .. attribute:: cw_pvt_size
        The sizing bounds for the clean-water PV-T collectors.

    .. attribute:: hot_water_tanks
        The sizing bounds for the hot-water tanks.

    .. attribute:: hw_pvt_size
        The sizing bounds for the hot-water PV-T collectors.

    .. attribute:: iteration_length
        The length of the iterations to be carried out.

    .. attribute:: number_of_iterations
        The number of iterations to carry out.

    .. attribute:: pv_size
        The sizing bounds for the PV panels.

    .. attribute:: storage_size
        The sizing bounds for the electricity storage system (i.e., batteries).

    """

    clean_water_tanks: TankSize
    converter_sizes: Dict[Converter, ConverterSize]
    cw_pvt_size: SolarSystemSize
    hot_water_tanks: TankSize
    hw_pvt_size: SolarSystemSize
    iteration_length: int
    number_of_iterations: int
    pv_size: SolarSystemSize
    storage_size: StorageSystemSize

    @classmethod
    def from_dict(  # pylint: disable=too-many-statements
        cls,
        available_converters: List[Converter],
        logger: Logger,
        optimisation_inputs: Dict[str, Any],
    ) -> Any:
        """
        Returns a :class:`OptimisationParameters` instance based on the input info.

        Inputs:
            - available_converters:
                The `list` of :class:`conversion.Converter` instances that are defined
                and which are available to the system.
            - logger:
                The :class:`logging.Loggger` to use for the run.
            - optimisation_inputs:
                The optimisation input information, extracted from the input file.

        Outputs:
            - A :class:`OptimisationParameters` instanced based on the information
            passed in.

        """

        # Parse the clean-water PV-T system size.
        if OptimisationComponent.CLEAN_WATER_PVT_SIZE.value in optimisation_inputs:
            try:
                cw_pvt_size = SolarSystemSize(
                    optimisation_inputs[
                        OptimisationComponent.CLEAN_WATER_PVT_SIZE.value
                    ][MAX],
                    optimisation_inputs[
                        OptimisationComponent.CLEAN_WATER_PVT_SIZE.value
                    ][MIN],
                    optimisation_inputs[
                        OptimisationComponent.CLEAN_WATER_PVT_SIZE.value
                    ][STEP],
                )
            except KeyError:
                logger.error(
                    "%sNot all clean-water PV-T size information specified in the "
                    "optimisation inputs file.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise
            if cw_pvt_size.min == 0 or cw_pvt_size.max == 0:
                logger.error(
                    "%sCannot have zero clean-water PV-T collectors when modelling the "
                    "clean-water system.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InputFileError(
                    "optimisation inputs",
                    "If modelling a clean-water system, none of the clean-water PV-T "
                    "size options can be set to zero.",
                )
        else:
            cw_pvt_size = SolarSystemSize()

        # Parse the clean-water tank information.
        if OptimisationComponent.CLEAN_WATER_TANKS.value in optimisation_inputs:
            try:
                clean_water_tanks: TankSize = TankSize(
                    int(
                        optimisation_inputs[
                            OptimisationComponent.CLEAN_WATER_TANKS.value
                        ][MAX]
                    ),
                    int(
                        optimisation_inputs[
                            OptimisationComponent.CLEAN_WATER_TANKS.value
                        ][MIN]
                    ),
                    int(
                        optimisation_inputs[
                            OptimisationComponent.CLEAN_WATER_TANKS.value
                        ][STEP]
                    ),
                )
            except KeyError:
                logger.error(
                    "%sNot all clean-water tank information specified in the "
                    "optimisation inputs file.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise
            if clean_water_tanks.min == 0 or clean_water_tanks.max == 0:
                logger.error(
                    "%sCannot have zero clean-water tanks when modelling the "
                    "clean-water system.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InputFileError(
                    "optimisation inputs",
                    "If modelling a clean-water system, none of the clean-water tank "
                    "size options can be set to zero.",
                )
        else:
            clean_water_tanks = TankSize()

        # Parse the converters that are to be optimised.
        converter_sizing_inputs: Dict[str, Dict[str, int]] = {
            key: value  # type: ignore
            for key, value in optimisation_inputs.items()
            if CONVERTER_SIZE_REGEX.match(key) is not None
        }

        # NOTE: Explicit error handling is done for the type-check ignored lines.
        try:
            converter_sizing_inputs = {
                CONVERTER_SIZE_REGEX.match(key).group(CONVERTER_NAME_STRING): value  # type: ignore
                for key, value in converter_sizing_inputs.items()
                if CONVERTER_SIZE_REGEX.match(key).group(CONVERTER_NAME_STRING)  # type: ignore
                in {converter.name for converter in available_converters}
            }
        except AttributeError:
            logger.error(
                "%sError parsing converter input information, unable to match groups."
                "%s",
                BColours.fail,
                BColours.endc,
            )
            raise

        converter_name_to_converter = {
            converter.name: converter for converter in available_converters
        }
        try:
            converter_sizes: Dict[Converter, ConverterSize] = {
                converter_name_to_converter[key]: ConverterSize(
                    entry[MAX], entry[MIN], entry[STEP]
                )
                for key, entry in converter_sizing_inputs.items()
            }
        except KeyError:
            logger.error(
                "%sNot all information was provided for the converters defined within "
                "the optimisation inputs file.%s",
                BColours.fail,
                BColours.endc,
            )
            raise

        # Parse the hot-water PV-T size information.
        if OptimisationComponent.HOT_WATER_PVT_SIZE.value in optimisation_inputs:
            try:
                hw_pvt_size = SolarSystemSize(
                    optimisation_inputs[OptimisationComponent.HOT_WATER_PVT_SIZE.value][
                        MAX
                    ],
                    optimisation_inputs[OptimisationComponent.HOT_WATER_PVT_SIZE.value][
                        MIN
                    ],
                    optimisation_inputs[OptimisationComponent.HOT_WATER_PVT_SIZE.value][
                        STEP
                    ],
                )
            except KeyError:
                logger.error(
                    "%sNot all hot-water PV-T size information specified in the "
                    "optimisation inputs file.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise
            if hw_pvt_size.min == 0 or hw_pvt_size.max == 0:
                logger.error(
                    "%sCannot have zero hot-water PV-T collectors when modelling the "
                    "hot-water system.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InputFileError(
                    "optimisation inputs",
                    "If modelling an hot-water system, none of the hot-water PV-T size "
                    "options can be set to zero.",
                )
        else:
            hw_pvt_size = SolarSystemSize()

        # Parse the hot-water tank information.
        if OptimisationComponent.HOT_WATER_TANKS.value in optimisation_inputs:
            try:
                hot_water_tanks: TankSize = TankSize(
                    int(
                        optimisation_inputs[
                            OptimisationComponent.HOT_WATER_TANKS.value
                        ][MAX]
                    ),
                    int(
                        optimisation_inputs[
                            OptimisationComponent.HOT_WATER_TANKS.value
                        ][MIN]
                    ),
                    int(
                        optimisation_inputs[
                            OptimisationComponent.HOT_WATER_TANKS.value
                        ][STEP]
                    ),
                )
            except KeyError:
                logger.error(
                    "%sNot all hot-water tank information specified in the "
                    "optimisation inputs file.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise
            if hot_water_tanks.min == 0 or hot_water_tanks.max == 0:
                logger.error(
                    "%sCannot have zero hot-water tanks when modelling the "
                    "hot-water system.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InputFileError(
                    "optimisation inputs",
                    "If modelling an hot-water system, none of the hot-water tank "
                    "size options can be set to zero.",
                )
        else:
            hot_water_tanks = TankSize()

        if OptimisationComponent.PV_SIZE.value in optimisation_inputs:
            try:
                pv_size = SolarSystemSize(
                    optimisation_inputs[OptimisationComponent.PV_SIZE.value][MAX],
                    optimisation_inputs[OptimisationComponent.PV_SIZE.value][MIN],
                    optimisation_inputs[OptimisationComponent.PV_SIZE.value][STEP],
                )
            except KeyError:
                logger.error(
                    "%sNot all PV size information specified in the optimisation "
                    "inputs file.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise
        else:
            pv_size = SolarSystemSize()

        if OptimisationComponent.STORAGE_SIZE.value in optimisation_inputs:
            try:
                storage_size = StorageSystemSize(
                    optimisation_inputs[OptimisationComponent.STORAGE_SIZE.value][MAX],
                    optimisation_inputs[OptimisationComponent.STORAGE_SIZE.value][MIN],
                    optimisation_inputs[OptimisationComponent.STORAGE_SIZE.value][STEP],
                )
            except KeyError:
                logger.error(
                    "%sNot all battery storage size information specified in the "
                    "optimisation inputs file.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise
        else:
            storage_size = StorageSystemSize()

        return cls(
            clean_water_tanks,
            converter_sizes,
            cw_pvt_size,
            hot_water_tanks,
            hw_pvt_size,
            optimisation_inputs[ITERATION_LENGTH],
            optimisation_inputs[NUMBER_OF_ITERATIONS],
            pv_size,
            storage_size,
        )

    @property
    def scenario_length(self) -> int:
        """
        Calculates and returns the scenario length for the optimisation.

        Outputs:
            - The scenario length for the optimisation.

        """

        return self.iteration_length * self.number_of_iterations

    def to_dict(self) -> Dict[str, Union[int, float]]:
        """
        Returns a `dict` representation of the :class:`OptimisationParameters` instance.

        Outputs:
            A `dict` containing the :class:`OptimisationParameters` information.

        """

        optimisation_parameters_dict = {
            "clean_water_pvt_size_max": int(self.cw_pvt_size.max)
            if self.cw_pvt_size is not None
            else None,
            "clean_water_pvt_size_min": int(self.cw_pvt_size.min)
            if self.cw_pvt_size is not None
            else None,
            "clean_water_pvt_size_step": int(self.cw_pvt_size.step)
            if self.cw_pvt_size is not None
            else None,
            "clean_water_tanks_max": int(self.clean_water_tanks.max)
            if self.clean_water_tanks is not None
            else None,
            "clean_water_tanks_min": int(self.clean_water_tanks.min)
            if self.clean_water_tanks is not None
            else None,
            "clean_water_tanks_step": int(self.clean_water_tanks.step)
            if self.clean_water_tanks is not None
            else None,
            "hot_water_pvt_size_max": int(self.hw_pvt_size.max)
            if self.hw_pvt_size is not None
            else None,
            "hot_water_pvt_size_min": int(self.hw_pvt_size.min)
            if self.hw_pvt_size is not None
            else None,
            "hot_water_pvt_size_step": int(self.hw_pvt_size.step)
            if self.hw_pvt_size is not None
            else None,
            "hot_water_tanks_max": int(self.hot_water_tanks.max)
            if self.hot_water_tanks is not None
            else None,
            "hot_water_tanks_min": int(self.hot_water_tanks.min)
            if self.hot_water_tanks is not None
            else None,
            "hot_water_tanks_step": int(self.hot_water_tanks.step)
            if self.hot_water_tanks is not None
            else None,
            ITERATION_LENGTH: round(self.iteration_length, 3),
            NUMBER_OF_ITERATIONS: round(self.number_of_iterations, 3),
            "pv_size_max": round(self.pv_size.max, 3),
            "pv_size_min": round(self.pv_size.min, 3),
            "pv_size_step": round(self.pv_size.step, 3),
            "storage_size_max": round(self.storage_size.max, 3),
            "storage_size_min": round(self.storage_size.min, 3),
            "storage_size_step": round(self.storage_size.step, 3),
        }

        return {
            key: value
            for key, value in optimisation_parameters_dict.items()
            if value is not None
        }


# Threshold-criterion-to-mode mapping:
#   Maps the threshold criteria to the modes, i.e., whether they are maximisable or
#   minimisable.
THRESHOLD_CRITERION_TO_MODE: Dict[Criterion, ThresholdMode] = {
    Criterion.BLACKOUTS: ThresholdMode.MAXIMUM,
    Criterion.CLEAN_WATER_BLACKOUTS: ThresholdMode.MAXIMUM,
    Criterion.CUMULATIVE_COST: ThresholdMode.MAXIMUM,
    Criterion.CUMULATIVE_GHGS: ThresholdMode.MAXIMUM,
    Criterion.CUMULATIVE_SYSTEM_COST: ThresholdMode.MAXIMUM,
    Criterion.EMISSIONS_INTENSITY: ThresholdMode.MAXIMUM,
    Criterion.KEROSENE_COST_MITIGATED: ThresholdMode.MINIMUM,
    Criterion.KEROSENE_DISPLACEMENT: ThresholdMode.MINIMUM,
    Criterion.KEROSENE_GHGS_MITIGATED: ThresholdMode.MINIMUM,
    Criterion.LCUE: ThresholdMode.MAXIMUM,
    Criterion.RENEWABLES_FRACTION: ThresholdMode.MINIMUM,
    Criterion.TOTAL_COST: ThresholdMode.MAXIMUM,
    Criterion.TOTAL_GHGS: ThresholdMode.MAXIMUM,
    Criterion.TOTAL_SYSTEM_COST: ThresholdMode.MAXIMUM,
    Criterion.TOTAL_SYSTEM_GHGS: ThresholdMode.MAXIMUM,
    Criterion.UNMET_ENERGY_FRACTION: ThresholdMode.MAXIMUM,
}


def get_sufficient_appraisals(
    optimisation: Optimisation, system_appraisals: List[SystemAppraisal]
) -> List[SystemAppraisal]:
    """
    Checks whether any of the system appraisals fulfill the threshold criterion

    Inputs:
        - optimisation:
            The optimisation currently being considered.
        - system_appraisals:
            Appraisals of the systems which have been simulated

    Outputs:
        - sufficient_systems:
            Appraisals of the systems which meet the threshold criterion (sufficient systems)

    """

    sufficient_appraisals: List[SystemAppraisal] = []

    # Cycle through the provided appraisals.
    for appraisal in system_appraisals:
        if appraisal.criteria is None:
            raise ProgrammerJudgementFault(
                "appraisal",
                "A system appraisal was returned which does not have criteria defined.",
            )
        criteria_met = set()
        for (
            threshold_criterion,
            threshold_value,
        ) in optimisation.threshold_criteria.items():
            # Add a `True` marker if the threshold criteria are met, otherwise add
            # False.
            if (
                THRESHOLD_CRITERION_TO_MODE[threshold_criterion]
                == ThresholdMode.MAXIMUM
            ):
                if appraisal.criteria[threshold_criterion] <= threshold_value:
                    criteria_met.add(True)
                else:
                    criteria_met.add(False)
            if (
                THRESHOLD_CRITERION_TO_MODE[threshold_criterion]
                == ThresholdMode.MINIMUM
            ):
                if appraisal.criteria[threshold_criterion] >= threshold_value:
                    criteria_met.add(True)
                else:
                    criteria_met.add(False)

        # Store the system to return provided it is sufficient.
        if all(criteria_met):
            sufficient_appraisals.append(appraisal)

    return sufficient_appraisals


def recursive_iteration(  # pylint: disable=too-many-locals
    conventional_cw_source_profiles: Optional[Dict[WaterSource, pd.DataFrame]],
    disable_tqdm: bool,
    end_year: int,
    finance_inputs: Dict[str, Any],
    ghg_inputs: Dict[str, Any],
    grid_profile: Optional[pd.DataFrame],
    irradiance_data: pd.Series,
    kerosene_usage: pd.DataFrame,
    location: Location,
    logger: Logger,
    minigrid: energy_system.Minigrid,
    optimisation: Optimisation,
    previous_system: Optional[SystemAppraisal],
    start_year: int,
    temperature_data: pd.Series,
    total_loads: Dict[ResourceType, Optional[pd.DataFrame]],
    total_solar_pv_power_produced: pd.Series,
    wind_speed_data: Optional[pd.Series],
    yearly_electric_load_statistics: pd.DataFrame,
    *,
    component_sizes: Dict[
        Union[Converter, ImpactingComponent, RenewableEnergySource],
        Union[int, float],
    ],
    parameter_space: List[
        Tuple[
            Union[Converter, ImpactingComponent, RenewableEnergySource],
            str,
            Union[List[int], List[float]],
        ]
    ],
    system_appraisals: List[SystemAppraisal],
) -> List[SystemAppraisal]:
    """
    Recursively look for sufficient systems through a series of parameter spaces.

    To recursively search through the parameter space, two objects are utilised:
    - a mapping between the component and the size to model for it;
    - a `list` of `tuple`s containing the components along with a list of possible
      values.

    At each stage in the process, a single component is removed from the `list` and
    added to the mapping such that a definite value is assigned. This is then passed
    through recursively with the function being called each time from a loop. In this
    way, a single level of the recursion deals with a single component of the system
    and its possible sizes, whilst the lowest (deepest) level of recursion deals with
    the actual simulations that are being carried out.

    In order to specify unique values, i.e., components of the system which have a fixed
    size and do not require iteration, use the mapping directly. In this way, the
    recursive function will be unaware of whether there exists a recursive layer for
    this parameter or whether the value has been uniquely defined.

    Inputs: (NOTE: Only inputs are listed which are utilised within this function.)
        - component_sizes:
            Specific values for the varoius :class:`finance.ImpactingComponent` sizes
            and :class:`RenewableEnergySource` sizes to use for the simulation to be
            carried out.
        - parameter_spaces:
            A `list` containing `tuple`s as entries that specify:
            - The :class:`finance.ImpactingComponent` that should have its sizes
              iterated through;
            - The unit to display to the user, as a `str`;
            - The `list` of values to iterate through.
        - system_appraisals:
            The `list` containing the :class:`SystemAppraisal` instances that correspond
            to sufficient systems.

    Outputs:
        - A `list` of sufficient systems.

    """

    # If there are no more things to iterate through, then run a simulation and return
    # whether the system was sufficient.
    if len(parameter_space) == 0:
        logger.info(
            "Running simulation with component sizes:\n- %s",
            "\n- ".join(
                [f"{key.value} size={value}" for key, value in component_sizes.items()]
            ),
        )

        # Determine the converter sizes.
        if not all(isinstance(value, int) for value in component_sizes.values()):
            logger.info(
                "%sNon-integer component sizes were specified.%s",
                BColours.fail,
                BColours.endc,
            )
        converters = converters_from_sizing(
            {
                key: int(value)
                for key, value in component_sizes.items()
                if isinstance(key, Converter)
            }
        )

        # Run the simulation
        (
            _,
            simulation_results,
            system_details,
        ) = energy_system.run_simulation(
            int(component_sizes[RenewableEnergySource.CLEAN_WATER_PVT]),
            conventional_cw_source_profiles,
            converters,
            disable_tqdm,
            component_sizes[ImpactingComponent.STORAGE],
            grid_profile,
            int(component_sizes[RenewableEnergySource.HOT_WATER_PVT]),
            irradiance_data,
            kerosene_usage,
            location,
            logger,
            minigrid,
            int(component_sizes[ImpactingComponent.CLEAN_WATER_TANK]),
            int(component_sizes[ImpactingComponent.HOT_WATER_TANK]),
            total_solar_pv_power_produced,
            component_sizes[RenewableEnergySource.PV],
            optimisation.scenario,
            Simulation(end_year, start_year),
            temperature_data,
            total_loads,
            wind_speed_data,
        )

        new_appraisal = appraise_system(
            yearly_electric_load_statistics,
            end_year,
            finance_inputs,
            ghg_inputs,
            location,
            logger,
            previous_system,
            optimisation.scenario,
            simulation_results,
            start_year,
            system_details,
        )

        return get_sufficient_appraisals(optimisation, [new_appraisal])

    # If there are things to iterate through, then iterate through these, calling the
    # function recursively.
    component, unit, sizes = parameter_space.pop()

    for size in tqdm(
        sizes,
        desc=f"{component.value} size options",
        disable=disable_tqdm,
        leave=False,
        unit=unit,
    ):
        # Update the set of fixed sizes accordingly.
        updated_component_sizes: Dict[
            Union[Converter, ImpactingComponent, RenewableEnergySource],
            Union[int, float],
        ] = component_sizes.copy()
        updated_component_sizes[component] = size

        # Call the function recursively.
        sufficient_appraisals = recursive_iteration(
            conventional_cw_source_profiles,
            disable_tqdm,
            end_year,
            finance_inputs,
            ghg_inputs,
            grid_profile,
            irradiance_data,
            kerosene_usage,
            location,
            logger,
            minigrid,
            optimisation,
            previous_system,
            start_year,
            temperature_data,
            total_loads,
            total_solar_pv_power_produced,
            wind_speed_data,
            yearly_electric_load_statistics,
            component_sizes=updated_component_sizes,
            parameter_space=parameter_space.copy(),
            system_appraisals=system_appraisals,
        )
        if sufficient_appraisals == []:
            logger.info("No sufficient systems at this resolution.")
            if len(parameter_space) == 0:
                logger.info("Probing lowest depth - skipping further size options.")
                break
            logger.info("Probing non-lowest depth - continuing iteration.")
            continue

        # Store the new appraisal if it is sufficient.
        logger.info("Sufficient system found, storing.")
        for appraisal in sufficient_appraisals:
            if appraisal.criteria is None:
                logger.error(
                    "%sNo appraisal criteria for appraisal.%s",
                    BColours.fail,
                    BColours.endc,
                )
                logger.debug("System appraisal: %s", appraisal)
                raise ProgrammerJudgementFault(
                    "appraisal module",
                    "When processing debug output for sufficient appraisals, an error "
                    "occured as there were no criteria attached to the appraisal. More "
                    "information can be found in the logger directory.",
                )
            logger.debug(
                "Threshold criteria: %s",
                json.dumps(
                    {
                        str(key.value): value
                        for key, value in appraisal.criteria.items()
                    },
                    indent=4,
                ),
            )
        system_appraisals.extend(sufficient_appraisals)

    # Return the sufficient appraisals that were found at this resolution.
    return sufficient_appraisals


def save_optimisation(
    disable_tqdm: bool,
    logger: Logger,
    optimisation: Optimisation,
    optimisation_inputs: OptimisationParameters,
    optimisation_number: int,
    output: str,
    output_directory: str,
    scenario: Scenario,
    system_appraisals: List[SystemAppraisal],
) -> None:
    """
    Saves simulation outputs to a .csv file

    Inputs:
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - logger:
            The logger to use for the run.
        - optimisation_inputs:
            The optimisation input information.
        - optimisation_number:
            The number of the optimisation that has just been carried out.
        - output:
            The output name to use when labelling the simulation: this is the name given
            to the output folder in which the system files are saved.
        - output_directory:
            The directory into which the files should be saved.
        - scenario:
            The scenario for the optimisation.
        - system_appraisals:
            A `list` of the :class:`SystemAppraisal` instances which specify the
            optimum systems at each time step.

    """

    # Remove the file extension if appropriate.
    if output.endswith(".json"):
        output = output.rsplit(".json", 1)[0]

    # Create the output directory.
    optimisation_output_folder = os.path.join(output_directory, output)
    os.makedirs(optimisation_output_folder, exist_ok=True)

    # Add the key results to the system data.
    system_appraisals_dict = {
        f"iteration_{index}": appraisal.to_dict()
        for index, appraisal in enumerate(system_appraisals)
    }

    # Add the optimisation parameter information.
    output_dict = {
        "optimisation_inputs": optimisation_inputs.to_dict(),
        "optimisation_criteria": {
            key.value: value.value
            for key, value in optimisation.optimisation_criteria.items()
        },
        "threshold_criteria": {
            key.value: value for key, value in optimisation.threshold_criteria.items()
        },
        "scenario": scenario.to_dict(),
        "system_appraisals": system_appraisals_dict,
    }

    with tqdm(
        total=1,
        desc="saving output files",
        disable=disable_tqdm,
        leave=False,
        unit="file",
    ) as pbar:
        # Save the optimisation data.
        logger.info("Saving optimisation output.")
        with open(
            os.path.join(
                optimisation_output_folder,
                f"optimisation_output_{optimisation_number}.json",
            ),
            "w",
        ) as f:
            json.dump(output_dict, f, indent=4)
        logger.info(
            "Optimisation successfully saved to %s.", optimisation_output_folder
        )
        pbar.update(1)



#!/usr/bin/python3
########################################################################################
# transmission.py - Transmission module.                                               #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2020                                                      #
# Date created: 08/11/2021                                                             #
# License: Open source                                                                 #
# Most recent update: 08/11/2021                                                       #
########################################################################################
"""
transmission.py - The transmission module for CLOVER.

CLOVER considers the transportation of different media around the minigrid. This module
is concerned with the modelling and consideration of these aspects of the energy system.

"""

import dataclasses

from logging import Logger
from typing import Any, Dict, List

from ..__utils__ import (
    BColours,
    InputFileError,
    NAME,
    RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING,
    ResourceType,
)

__all__ = ("Transmitter",)


# Maximum throughput:
#   The maximum throughput through the transmitter.
MAXIMUM_THROUGHPUT: str = "maximum_throughput"

# Transmits:
#   Keyword used for parsing the material transmitted by the transmitter.
TRANSMITS: str = "transmits"


@dataclasses.dataclass
class Transmitter:
    """
    Represents a transmitter of a resource within the minigrid system.

    .. attribute:: consumption
        The consumption of the :class:`Transmitter`.

    .. attribute:: name
        The name of the :class:`Transmitter`.

    .. attribute:: power_source_material
        The :class:`ResourceType` which is consumed to power the :class:`Tranismitter`.

    .. attribute:: throughput
        The maximum throughput of the :class:`Transmitter`.

    .. attribute:: transmission_material
        The :class:`ResourceType` being conveyed by the transmitter.

    """

    consumption: float
    name: str
    power_source_material: ResourceType
    throughput: float
    transmission_material: ResourceType

    def __hash__(self) -> int:
        """
        Return a unique identifier for the device.

        Outputs:
            A unique identifier for the device.

        """

        return hash(self.__str__())

    def __str__(self) -> str:
        """
        Return a nice-looking output for the device.

        Outputs:
            - A nice-looking string representation for the device.

        """

        representation_string = (
            "Transmitter("
            + f"name={self.name} transmitting {self.transmission_material}, "
            + f"conusming {self.consumption} {self.power_source_material}, "
            + f"throughput={self.throughput}"
            + ")"
        )

        return representation_string

    @classmethod
    def from_dict(cls, input_data: Dict[str, Any], logger: Logger) -> Any:
        """
        Processes input data to generate a :class:`Transmitter` instance.

        Inputs:
            - input_data:
                The transmitter input data extracted from the transmission input file.

        Outputs:
            - The :class:`Transmitter` instancce based on the input data.

        """

        # Determine the input load type.
        input_resource_list: List[str] = [
            str(key)
            for key in input_data
            if key in RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING
        ]
        if len(input_resource_list) > 1:
            logger.error(
                "%sTransmitters can only be powered by a single resource.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "transmission inputs",
                "Transmitters can only be powered by a single resource.",
            )

        # Determine the consumption of this resource.
        input_resource_consumption: Dict[ResourceType, float] = {}
        for input_resource in input_resource_list:
            try:
                input_resource_consumption[
                    ResourceType(RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING[input_resource])
                ] = float(input_data[input_resource])
            except TypeError as e:
                logger.error(
                    "%sInvalid entry in transmission file, check all value types are "
                    "correct: %s%s",
                    BColours.fail,
                    str(e),
                    BColours.endc,
                )
                raise InputFileError(
                    "transmission inputs",
                    f"Invalid value type in transmission file: {str(e)}",
                ) from None

        # Determine the single resource type allowed for the transmitter.
        input_resource = input_resource_list[0]

        consumption = input_resource_consumption[
            RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING[input_resource]
        ]

        return cls(
            consumption,
            input_data[NAME],
            ResourceType(RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING[input_resource]),
            input_data[MAXIMUM_THROUGHPUT],
            input_data[TRANSMITS],
        )



#!/usr/bin/python3
########################################################################################
# wind.py - Wind generation module  .                                                  #
#                                                                                      #
# Author: Phil Sandwell                                                                #
# Copyright: Phil Sandwell, 2021                                                       #
# License: Open source                                                                 #
# Most recent update: 14/07/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
wind.py - The wind-profile-generation module for CLOVER.

This module fetches wind profiles from renewables.ninja, parses them and saves them for
use locally within CLOVER.

"""

import enum

from typing import Any, Dict

import pandas as pd  # pylint: disable=import-error

from ..__utils__ import Location
from .__utils__ import BaseRenewablesNinjaThread, total_profile_output

__all__ = (
    "WindDataThread",
    "WIND_LOGGER_NAME",
    "total_wind_output",
)


# Wind logger name:
#   The name to use for the wind logger.
WIND_LOGGER_NAME = "wind_generation"


class WindDataThread(
    BaseRenewablesNinjaThread, profile_name="wind", profile_key="wind"
):
    """
    Class to use when calling the wind data thread.

    NOTE: Heights of wind turbines are limited to between 10 m and 150 m by the
    Renewables.Ninja API. This should be checked before attempting to fetch data beyond
    this range.

    """

    def __init__(
        self,
        auto_generated_files_directory: str,
        generation_inputs: Dict[str, Any],
        location: Location,
        logger_name: str,
        regenerate: bool,
        sleep_multiplier: int = 1,
        verbose: bool = False,
    ):
        """
        Instantiate a :class:`WindDataThread` instance.

        """

        # Add the additional parameters which are need when calling the wind data.
        renewables_ninja_params = {
            "lat": float(location.latitude),
            "lon": float(location.longitude),
            "local_time": "false",
            "capacity": 1.0,
            "height": 10,
            "turbine": "Vestas V80 2000",
            "raw": "true",
        }
        super().__init__(
            auto_generated_files_directory,
            generation_inputs,
            location,
            logger_name,
            regenerate,
            sleep_multiplier,
            verbose,
            renewables_ninja_params=renewables_ninja_params,
        )


def total_wind_output(*args, **kwargs) -> pd.DataFrame:
    """
    Wrapper function to wrap the total wind output.

    """

    return total_profile_output(*args, **kwargs, profile_name="wind")


class WindDataType(enum.Enum):
    """
    Stores weather condition information for extracting data from renewables ninja.

    - ELECTRICITY:
        Denotes the electricity generated by the wind turbine.

    - WIND_SPEED:
        Denotes the wind speed at the height specified for the data.

    """

    ELECTRICITY = "electricity"
    WIND_SPEED = "wind_speed"



#!/usr/bin/python3
########################################################################################
# update_api_token.py - Script for updating API tokens in locations.                   #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 21/09/2021                                                             #
#                                                                                      #
# For more information, please email:                                                  #
#   philip.sandwell@gmail.com                                                          #
########################################################################################
"""
update_api_token.py - Script for updating API tokens within locations.

This script, when run, scans through locations in search of API tokens to update.

"""

import argparse
import os
import sys

from typing import Any, Dict, List

import yaml  # pylint: disable=import-error

from ..__utils__ import get_logger, read_yaml
from ..fileparser import GENERATION_INPUTS_FILE, INPUTS_DIRECTORY, LOCATIONS_FOLDER_NAME
from ..generation.__utils__ import TOKEN

__all__ = ("main",)

# Logger name:
#   The name to use for the logger.
LOGGER_NAME: str = "update_api_token"


def _parse_args(args: List[Any]) -> argparse.Namespace:
    """
    Parse the CLI arguments to determine the flow of the script.

    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--token",
        type=str,
        help="The renewables.ninja API token to use.",
    )
    parser.add_argument(
        "--location", type=str, help="The name of the location to update."
    )

    return parser.parse_args(args)


def main(args: List[Any]) -> None:
    """
    The main method of the update api token script.

    Inputs:
        - args:
            The un-parsed command-line arguments.

    """

    logger = get_logger(LOGGER_NAME)
    logger.info("Update API token script called with arguments: %s", args)

    # Parse the command-line arguments.
    parsed_args = _parse_args(args)
    logger.info("Commane-line arguments successfully parsed.")

    # Error if not all arguments passed in.
    if parsed_args.location is None:
        raise Exception("A location to update must be specified.")
    if parsed_args.token is None:
        raise Exception("The renewables.ninja API token must be specifeid.")

    # Check whether the generation file exists.
    generation_file_path = os.path.join(
        LOCATIONS_FOLDER_NAME,
        parsed_args.location,
        INPUTS_DIRECTORY,
        GENERATION_INPUTS_FILE,
    )

    if not os.path.isfile(generation_file_path):
        raise FileNotFoundError(
            "The generation inputs file could not be found within the location "
            f"{parsed_args.location}.",
        )

    # Attempt to update the token.
    filedata: Dict[str, Any]
    try:
        filedata = read_yaml(generation_file_path, logger)  # type: ignore
    except Exception:
        logger.error("Error reading generation inputs file '%s'.", generation_file_path)
        raise

    filedata[TOKEN] = parsed_args.token

    logger.info("Attempting to save updated token to the generation inputs file.")
    with open(generation_file_path, "w") as f:
        yaml.dump(filedata, f)
    logger.info("Updated API token successfully saved.")


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/python3
########################################################################################
# solar.py - Solar panel modelling code for CLOVER.                                    #
#                                                                                      #
# Authors: Ben Winchester                                                              #
# Copyright: Phil Sandwell, 2021                                                       #
# Date created: 12/08/2021                                                             #
# License: Open source                                                                 #

# For more information, please email:                                                  #
#   philip.sandwell@gmail.com                                                          #
########################################################################################
"""
solar.py - Solar-panel modelling code for CLOVER.

In order to accurately model a solar panel within CLOVER, various information about its
performance under environmental conditions needs to be calculated.

"""

import collections

from logging import Logger
from typing import Dict, Optional, Tuple

import pandas as pd  # pylint: disable=import-error

from scipy import linalg
from tqdm import tqdm

from ..__utils__ import (
    BColours,
    HTFMode,
    InputFileError,
    InternalError,
    ResourceType,
    Scenario,
    ZERO_CELCIUS_OFFSET,
    dict_to_dataframe,
)
from ..conversion.conversion import ThermalDesalinationPlant
from .__utils__ import Minigrid
from .storage_utils import HotWaterTank


__all__ = ("calculate_pvt_output",)


# Temperature precision:
#   The precision required when solving the differential equation for the system
#   temperatures.
TEMPERATURE_PRECISION: float = 1.44


class MassFlowRateTooLargeError(Exception):
    """
    Raise when the mass-flow rate being specified is too large for the system to cope.

    """

    def __init__(self, msg: str) -> None:
        """
        Instantiate a :class:`MassFlowRateTooLargeError` instance.

        Inputs:
            - msg:
                The message to append.

        """

        super().__init__(
            f"The mass-flow rate was too large for the system to cope with: {msg}"
        )


class MassFlowRateTooSmallError(Exception):
    """
    Raise when the mass-flow rate being specified is too small for the system to cope.

    """

    def __init__(self, msg: str) -> None:
        """
        Instantiate a :class:`MassFlowRateTooSmallError` instance.

        Inputs:
            - msg:
                The message to append.

        """

        super().__init__(
            f"The mass-flow rate was too small for the system to operate: {msg}"
        )


def _htf_fed_buffer_tank_mass_flow_rate(
    ambient_temperature: float,
    best_guess_tank_temperature: float,
    buffer_tank: HotWaterTank,
    thermal_desalination_plant: ThermalDesalinationPlant,
) -> float:
    """
    Computes the mass-flow rate of HTF from the buffer tanks to the desalination plant.

    Inputs:
        - ambient_temperature:
            The ambient temperature, used as the base against which heat is being
            supplied. This is realistic as the final stage of desalination plants can be
            no less than the ambient temprature. This is measured in degrees Celcius.
        - best_guess_tank_temperature:
            The best guess at the tank temperature at the current time step, measured in
            degrees Celcius. This is the temperature at which HTF is removed from the
            tank to supply the desalination plant.
        - buffer_tank:
            The HTF hot-water tank.
        - thermal_desalination_plant:
            The thermal desalination plant being modelled.

    Outputs:
        - The mass-flow rate of HTF from the bugger tanks to the desalination plant,
          measured in litres per hour.

    Raises:
        - InputFileError:
            Raised if the thermal desalination plant does not use heat from the HTF.

    """

    if ResourceType.HEAT not in thermal_desalination_plant.input_resource_consumption:
        raise InputFileError(
            "converter inputs",
            "The thermal desalination plant selected does define its heat consumption.",
        )

    return (
        thermal_desalination_plant.input_resource_consumption[ResourceType.HEAT]
        * 3600  # [s/h]
        / (  # [Wth]
            buffer_tank.heat_capacity  # [J/kg*K]
            * (best_guess_tank_temperature - ambient_temperature)  # [K]
        )
    )  # [kg/hour]


def _volume_withdrawn_from_tank(
    ambient_temperature: float,
    best_guess_tank_temperature: float,
    hot_water_load: Optional[float],
    logger: Logger,
    minigrid: Minigrid,
    previous_tank_temperature: Optional[float],
    resource_type: ResourceType,
    thermal_desalination_plant: Optional[ThermalDesalinationPlant],
) -> Tuple[bool, float]:
    """
    Computes whether the tank is supplying an output, and what this output is.

    Inputs:


    Outputs:
        A `tuple` containing:
        - tank_supply_on:
            Whether liquid was withdrawn from the tank (True) or not (False).
        - volume_supplied:
            The volume supplied, measured in kg/hour.

    """

    if resource_type == ResourceType.CLEAN_WATER:
        if previous_tank_temperature is None or thermal_desalination_plant is None:
            logger.error(
                "%sNot enough parameters specified to determine buffer-tank mass-flow "
                "rate.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "Not enough desalination-specific parameters specified to determine "
                "the buffer-tank mass-flow rate."
            )
        if minigrid.buffer_tank is None:
            logger.error(
                "%sNo buffer tank specified when attempting to determine the volume "
                "withdrawn from the tank.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "No buffer tank was found defined for the minigrid specified despite "
                "clean-water loads being requested."
            )

        # If the plant is heated by HTF.
        if thermal_desalination_plant.htf_mode == HTFMode.CLOSED_HTF:
            if thermal_desalination_plant.minimum_htf_temperature is None:
                logger.error(
                    "%sNo minimum htf temperature defined despite '%s' being the HTF "
                    "mode.%s",
                    BColours.fail,
                    thermal_desalination_plant.htf_mode.value,
                    BColours.endc,
                )
                raise InternalError(
                    "Minimum HTF temperature unexpectly undefined when computing plant "
                    "supply volume."
                )

            tank_supply_on: bool = (
                previous_tank_temperature
                > thermal_desalination_plant.minimum_htf_temperature
            )
            volume_supplied: float = (
                _htf_fed_buffer_tank_mass_flow_rate(
                    ambient_temperature,
                    best_guess_tank_temperature,
                    minigrid.buffer_tank,
                    thermal_desalination_plant,
                )
                * tank_supply_on
            )
        if thermal_desalination_plant.htf_mode == HTFMode.FEEDWATER_HEATING:
            if thermal_desalination_plant.minimum_feedwater_temperature is None:
                logger.error(
                    "%sNo minimum feedwater temperature defined despite '%s' being the "
                    "HTF mode.%s",
                    BColours.fail,
                    thermal_desalination_plant.htf_mode.value,
                    BColours.endc,
                )
                raise InternalError(
                    "Minimum feedwater temperature unexpectly undefined when computing "
                    "plant supply volume."
                )

            tank_supply_on = (
                previous_tank_temperature
                > thermal_desalination_plant.minimum_feedwater_temperature
            )
            volume_supplied = (
                thermal_desalination_plant.input_resource_consumption[
                    ResourceType.HOT_UNCLEAN_WATER
                ]
                * tank_supply_on
            )

    elif resource_type == ResourceType.HOT_CLEAN_WATER:
        if hot_water_load is None:
            logger.error(
                "%sNo hot-water load defined despite a hot-water system being "
                "specified.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "No hot-water load was provided to the 'tank volume supplied' "
                "calculation method despite hot water being defined in the scenario "
                "file."
            )
        tank_supply_on = hot_water_load > 0
        volume_supplied = hot_water_load

    return tank_supply_on, volume_supplied


def calculate_pvt_output(  # pylint: disable=too-many-locals, too-many-statements
    disable_tqdm: bool,
    end_hour: int,
    irradiances: pd.Series,
    logger: Logger,
    minigrid: Minigrid,
    num_tanks: int,
    processed_total_hw_load: Optional[pd.Series],
    pvt_system_size: int,
    resource_type: ResourceType,
    scenario: Scenario,
    start_hour: int,
    temperatures: pd.Series,
    thermal_desalination_plant: Optional[ThermalDesalinationPlant],
    wind_speeds: pd.Series,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Computes the output of a PV-T system.

    Inputs:
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - end_hour:
            The end hour for the simulation being carried out.
        - irradiances:
            The :class:`pd.Series` containing irradiance information for the time
            period being modelled.
        - logger:
            The logger to use for the run.
        - minigrid:
            The minigrid being modelled currently.
        - num_tanks:
            The number of hot-water tanks being modelled currently, which can either be
            buffer tanks (for desalination systems), or hot-water tanks (for hot-water
            systems).
        - processed_total_hw_load:
            The total hot-water load placed on the system, measured in litres per hour.
        - pvt_system_size:
            The size of the PV-T system being modelled.
        - resource_type:
            The resource type for which the PV-T output is being determined.
        - scenario:
            The :class:`Scenario` being considered.
        - start_hour:
            The start hour for the simulation being carried out.
        - temperatures:
            The :class:`pd.Series` containing temperature information for the time
            period being modelled.
        - thermal_desalination_plant:
            The thermal desalination plant being considered.
        - wind_speeds:
            The :class:`pd.Series` containing wind-speed information for the time period
            being modelled.

    Outputs:
        - pvt_collector_output_temperature:
            The output temperature of the PV-T collectors at each time step.
        - pvt_electric_power_per_unit:
            The electric power, per unit PV-T, delivered by the PV-T system.
        - pvt_pump_times_frame:
            The times for which the PV-T pump was switched on.
        - tank_temperature:
            The tank temperatures throughout the simulation, measured in degrees C.
        - tank_volume_supplied:
            The amount of hot water supplied by the hot-water tanks to the desalination
            system.

    """

    if minigrid.pvt_panel is None:
        logger.error(
            "%sThe energy system does not contain a PV-T panel despite the PV-T output "
            "computation function being called.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "energy system inputs",
            "The energy system specified does not contain a PV-T panel but PV-T "
            "modelling was requested.",
        )
    if minigrid.heat_exchanger is None:
        logger.error(
            "%sThe energy system does not contain a heat exchanger despite the PV-T "
            "output computation function being called which is reliant on the "
            "definition of a heat exchanger.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "energy system inputs",
            "The energy system specified does not contain a heat exchanger but PV-T "
            "modelling was requested for which this is required.",
        )

    # Instantiate debugging variables
    runs: int = 0

    # Instantiate maps for easy PV-T power lookups.
    pvt_electric_power_per_unit_map: Dict[int, float] = {}
    pvt_pump_times_map: Dict[int, int] = {}
    tank_supply_temperature_map: Dict[  # pylint: disable=unused-variable
        int, float
    ] = {}
    tank_volume_supplied_map: Dict[int, float] = {}

    # Compute the various terms which remain common across all time steps.
    if (
        scenario.desalination_scenario is not None
        and resource_type == ResourceType.CLEAN_WATER
    ):
        if minigrid.buffer_tank is None:
            logger.error(
                "%sThe energy system does not contain a buffer tank despite the PV-T "
                "output computation function being called for a clean-water load.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "energy system inputs",
                "The energy system specified does not contain a buffer tank but PV-T "
                "modelling for desalination was requested.",
            )

        default_supply_temperature: float = (
            scenario.desalination_scenario.feedwater_supply_temperature
        )
        mass_flow_rate: float = (
            scenario.desalination_scenario.pvt_scenario.mass_flow_rate
        )
        tank: HotWaterTank = minigrid.buffer_tank

        pvt_heat_transfer: float = (
            pvt_system_size
            * mass_flow_rate  # [kg/hour]
            * scenario.desalination_scenario.pvt_scenario.htf_heat_capacity  # [J/kg*K]
            * minigrid.heat_exchanger.efficiency
            / 3600  # [s/hour]
        )  # [W/K]
        tank_replacement_temperature: float = default_supply_temperature  # [degC]

        # Throw an error if the PV-T is not heating an intermediary HTF.
        if scenario.desalination_scenario.pvt_scenario.heats != HTFMode.CLOSED_HTF:
            logger.error(
                "%sCurrently, closed HTF PV-T modelling is supported only.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "desalination scenario",
                "The PV-T heating mode requested is not supported.",
            )

    elif (
        scenario.hot_water_scenario is not None
        and resource_type == ResourceType.HOT_CLEAN_WATER
    ):
        if minigrid.hot_water_tank is None:
            logger.error(
                "%sThe energy system does not contain a hot-water tank despite the "
                "PV-T output computation function being called for a hot-water load.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "energy system inputs",
                "The energy system specified does not contain a hot-water tank but "
                "PV-T modelling for hot-water loads was requested.",
            )

        default_supply_temperature = (
            scenario.hot_water_scenario.cold_water_supply_temperature
        )
        mass_flow_rate = scenario.hot_water_scenario.pvt_scenario.mass_flow_rate
        tank = minigrid.hot_water_tank

        pvt_heat_transfer = (
            pvt_system_size
            * mass_flow_rate  # [kg/hour]
            * scenario.hot_water_scenario.pvt_scenario.htf_heat_capacity  # [J/kg*K]
            * minigrid.heat_exchanger.efficiency
            / 3600  # [s/hour]
        )  # [W/K]
        tank_replacement_temperature = default_supply_temperature  # [degC]

    # One of these scenarios must be specified, so throw an error if not.
    else:
        logger.error(
            "%sNeither desalination nor hot-water scenarios were defined despite PV-T "
            "output being requested.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InternalError(
            "Neither desalination nor hot-water scenarios were specified despite PV-T "
            "being called."
        )

    best_guess_collector_input_temperature: float = default_supply_temperature
    pvt_collector_output_temperature_map: Dict[int, float] = collections.defaultdict(
        lambda: default_supply_temperature
    )
    tank_environment_heat_transfer: float = (
        num_tanks * tank.heat_transfer_coefficient
    )  # [W/K]
    tank_internal_energy: float = (
        num_tanks
        * tank.mass
        * tank.heat_capacity
        / 3600  # [kg]  # [J/kg*K]  # [s/hour]
    )  # [W/K]
    tank_temperature_map: Dict[int, float] = collections.defaultdict(
        lambda: default_supply_temperature
    )

    logger.info(
        "Beggining hourly %s PV-T performance calculation.", resource_type.value
    )
    for index in tqdm(
        range(start_hour, end_hour),
        desc=f"{resource_type.value.replace('_', ' ')} pv-t performance",
        disable=disable_tqdm,
        leave=False,
        unit="hour",
    ):
        # Determine whether the PV-T is flowing.
        if index > start_hour:
            pvt_flow_on: bool = (
                pvt_collector_output_temperature_map[index - 1]
                > tank_temperature_map[index - 1]
            ) and irradiances[index] > 0
        else:
            pvt_flow_on = False

        previous_tank_temperature: float = (
            tank_temperature_map[index - 1]
            if index > start_hour
            else default_supply_temperature
        )

        # Determine the volume withdrawn from the buffer tanks
        tank_supply_on, volume_supplied = _volume_withdrawn_from_tank(
            temperatures[index],
            previous_tank_temperature,
            processed_total_hw_load[index]
            if processed_total_hw_load is not None
            else None,
            logger,
            minigrid,
            previous_tank_temperature,
            resource_type,
            thermal_desalination_plant,
        )

        # Only compute outputs if there is input irradiance.
        solution_found: bool = False
        # Keep processing until the temperatures are consistent.
        while not solution_found:
            # Use the AI to determine the output temperature of the collector, based on
            # the best guess of the collector input temperature.
            if irradiances[index] > 0:
                (
                    fractional_electric_performance,
                    collector_output_temperature,
                ) = minigrid.pvt_panel.calculate_performance(
                    temperatures[index],
                    best_guess_collector_input_temperature,
                    logger,
                    mass_flow_rate,
                    1000 * irradiances[index],
                    wind_speeds[index],
                )
            else:
                fractional_electric_performance = 0
                collector_output_temperature = default_supply_temperature

            tank_load_enthalpy_transfer = (
                volume_supplied  # [kg/hour]
                * tank.heat_capacity  # [J/kg*K]
                / 3600  # [s/hour]
            )  # [W/K]

            # Determine the tank temperature and collector input temperature that match.
            resultant_vector = [
                (
                    pvt_heat_transfer
                    * (collector_output_temperature + ZERO_CELCIUS_OFFSET)
                    if pvt_flow_on
                    else 0
                )
                + tank_environment_heat_transfer
                * (temperatures[index] + ZERO_CELCIUS_OFFSET)
                + tank_internal_energy
                * (previous_tank_temperature + ZERO_CELCIUS_OFFSET)
                + (
                    tank_load_enthalpy_transfer
                    * (tank_replacement_temperature + ZERO_CELCIUS_OFFSET)
                    if tank_supply_on
                    else 0
                ),
                (
                    (1 - minigrid.heat_exchanger.efficiency)
                    * (collector_output_temperature + ZERO_CELCIUS_OFFSET)
                ),
            ]

            matrix = [
                [
                    0,
                    (
                        (pvt_heat_transfer if pvt_flow_on else 0)
                        + tank_environment_heat_transfer
                        + tank_internal_energy
                        + (tank_load_enthalpy_transfer if tank_supply_on else 0)
                    ),
                ],
                [1, -minigrid.heat_exchanger.efficiency],
            ]

            collector_input_temperature, tank_temperature = linalg.solve(
                a=matrix, b=resultant_vector
            )

            # Convert into Celcius.
            collector_input_temperature -= ZERO_CELCIUS_OFFSET
            tank_temperature -= ZERO_CELCIUS_OFFSET

            # If a solution has been found, then break the loop.
            if (
                abs(
                    collector_input_temperature - best_guess_collector_input_temperature
                )
                < TEMPERATURE_PRECISION
            ):
                runs = 0
                solution_found = True

            runs += 1
            if runs > 10:
                logger.debug(
                    "Index: %s: Run # %s: PV-T=%s, Plant=%s: Solution not yet found, "
                    "re-iterating: T_c,in=%s degC, best-guess T_c,in=%s degC, "
                    "T_tank=%s degC",
                    index,
                    runs,
                    "on" if pvt_flow_on else "off",
                    "on" if tank_supply_on else "off",
                    round(collector_input_temperature, 3),
                    round(best_guess_collector_input_temperature, 3),
                    round(tank_temperature, 3),
                )
            best_guess_collector_input_temperature = collector_input_temperature

        # Save the fractional electrical performance and output temp.
        pvt_collector_output_temperature_map[index] = collector_output_temperature
        pvt_electric_power_per_unit_map[index] = (
            fractional_electric_performance * minigrid.pvt_panel.pv_unit
        )
        pvt_pump_times_map[index] = int(pvt_flow_on)
        tank_temperature_map[index] = tank_temperature
        tank_volume_supplied_map[index] = volume_supplied

    logger.info("Hourly %s PV-T performance calculation complete.", resource_type.value)

    # Convert these outputs to dataframes and return.
    pvt_collector_output_temperature: pd.DataFrame = dict_to_dataframe(
        pvt_collector_output_temperature_map, logger
    )
    pvt_electric_power_per_unit: pd.DataFrame = dict_to_dataframe(
        pvt_electric_power_per_unit_map, logger
    )
    pvt_pump_times_frame: pd.DataFrame = dict_to_dataframe(pvt_pump_times_map, logger)
    tank_temperature_frame: pd.DataFrame = dict_to_dataframe(
        tank_temperature_map, logger
    )
    tank_volume_output_supplied: pd.DataFrame = dict_to_dataframe(
        tank_volume_supplied_map, logger
    )

    return (
        pvt_collector_output_temperature,
        pvt_electric_power_per_unit,
        pvt_pump_times_frame,
        tank_temperature_frame,
        tank_volume_output_supplied,
    )



#!/usr/bin/python3
########################################################################################
# __init__.py - Python internals module, no functionality here.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
# Date created: 05/07/2021
# License: Open source
########################################################################################
"""
simulation.__init__.py - Init module for the simulation component.

The `__init__` module ensures that all of the packages are correctly exposed so that
they can be imported when CLOVER is installed as a package.

"""

from .__utils__ import check_scenario, determine_available_converters
from .diesel import (
    DIESEL_CONSUMPTION,
    DieselGenerator,
    DieselWaterHeater,
    get_diesel_energy_and_times,
    get_diesel_fuel_usage,
)
from .energy_system import Minigrid, run_simulation
from .exchanger import Exchanger
from .solar import calculate_pvt_output
from .storage_utils import Battery, CleanWaterTank, HotWaterTank
from .storage import (
    battery_iteration_step,
    cw_tank_iteration_step,
    get_electric_battery_storage_profile,
    get_water_storage_profile,
)
from .transmission import Transmitter



#!/usr/bin/python3
########################################################################################
# hpc_outbox_assembly.py - CLOVER optimisation outbox assembly script.                 #
#                                                                                      #
# Authors: Ben Winchester                                                              #
# Copyright: Ben Winchester, 2022                                                      #
# Date created: 13/07/2021                                                             #
# License: Open source                                                                 #
#                                                                                      #
# For more information, please email:                                                  #
#   benedict.winchester@gmail.com                                                      #
########################################################################################
"""
hpc_outbox_assembly.py - CLOVER optimisation outbox assembly script.

This script assembles output files from optimisations that were carried out on Imperial
College London's High-Performance Computer(s) (HPC) into a single "outbox directory",
from where they can be easily copied from the HPC for futher analysis.

"""

import argparse
import os
import shutil
import subprocess
import sys

from typing import Any, List

# Destination directory:
#   THe name of the outbox directory into which to save optimisation output files.
DESTINATION_DIRECTORY: str = "optimisation_outbox"

# Optimisation name:
#   The name to use when finding optimisation output files.
OPTIMISATION_NAME: str = "optimisation_output_*.json"

# Simulation name:
#   The name to use when finding simulation output files.
SIMULATION_NAME: str = "simulation_output_*.csv"


def _parse_outbox_args(args: List[Any]) -> argparse.Namespace:
    """
    Parses the input arguments to the hpc outbox assembly script to determine the flow.

    Inputs:
        - args:
            Unparsed command-line arguments.

    Outputs:
        - The parsed HPC arguments.

    """

    parser = argparse.ArgumentParser()

    # Optimisation outputs:
    #   Whether to assemble optimisation outputs.
    parser.add_argument(
        "--optimisation",
        "-opt",
        action="store_true",
        default=False,
        help="Whether to include optimisation output files.",
    )

    # Optimisation outputs:
    #   Whether to assemble optimisation outputs.
    parser.add_argument(
        "--simulation",
        "-sim",
        action="store_true",
        default=False,
        help="Whether to include simulation output files.",
    )

    return parser.parse_args(args)


def main(args: List[Any]) -> None:
    """Main entry point for the CLOVER HPC outbox assembly script."""

    # Parse the command-line arguments.
    parsed_args = _parse_outbox_args(args)

    # Raise an error if no assembly was requested.
    if not parsed_args.simulation or parsed_args.optimisation:
        raise Exception(
            "Outbox assembly script must be called to either assemble optimisation or "
            "simulation output files or both. Neither argument was specified. Run with "
            "`--help` for more information."
        )

    filenames: List[str] = []

    if parsed_args.optimisation:
        # Determine the output files that have been generated by running an optimisation.
        optimisation_filenames = subprocess.run(
            ["find", "locations", "-name", OPTIMISATION_NAME],
            stdout=subprocess.PIPE,
            check=True,
        ).stdout.decode()
        optimisation_filenames_list = optimisation_filenames.split("\n")
        optimisation_filenames_list.remove("")
        filenames += optimisation_filenames_list

    if parsed_args.simulation:
        # Determine the output files that have been generated by running an optimisation.
        simulation_filenames = subprocess.run(
            ["find", "locations", "-name", SIMULATION_NAME],
            stdout=subprocess.PIPE,
            check=True,
        ).stdout.decode()
        simulation_filenames_list = simulation_filenames.split("\n")
        simulation_filenames_list.remove("")
        filenames += simulation_filenames_list

    # Copy each of these files to the outbox directory, including their directory tree
    # structure.
    for source_filename in filenames:
        destination_filename = os.path.join(DESTINATION_DIRECTORY, source_filename)

        # Make the directory structure within the outbox if it doesn't already exist
        if not os.path.isdir(os.path.dirname(destination_filename)):
            os.makedirs(os.path.dirname(destination_filename))

        # Copy across our file
        shutil.copy2(source_filename, destination_filename)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/python3
########################################################################################
# __utils__.py - CLOVER Impact Utility module.                                         #
#                                                                                      #
# Author: Phil Sandwell, Ben Winchester                                                #
# Copyright: Phil Sandwell, 2018                                                       #
# Date created: 24/08/2021                                                             #
#                                                                                      #
# For more information, please email:                                                  #
#   philip.sandwell@gmail.com                                                          #
########################################################################################
"""
__utils__.py - Utility module for CLOVER's Impact Component.

The utility module contains functionality which is used by the impact component modules,
namely the GHG and finance impact modules.

"""

import enum

from ..__utils__ import EXCHANGER

__all__ = (
    "ImpactingComponent",
    "LIFETIME",
    "WasteProduct",
)

# Lifetime:
#   Keyword to denote the lifetime of a component.
LIFETIME = "lifetime"

# Size increment:
#   Keyword to denote increments in the size of a component.
SIZE_INCREMENT = "size_increment"


class ImpactingComponent(enum.Enum):
    """
    Used to keep tracek of components within the systems that have associated impacts.

    - BOS:
        Denotes the balance-of-systems aspect of the system.

    - BUFFER_TANK:
        Denotes the buffer tank component of the system.

    - CLEAN_WATER_TANK:
        Denotes the clean-water tank component of the system.

    - CONVENTIONAL_SOURCE:
        Denotes conventional sources which would be utilised during minigrid supply
        downtimes.

    - CONVERTER:
        Denotes a component of the system responsible for the conversion of one
        resource type into another.

    - DIESEL:
        Denotes the diesel component of the system.

    - DIESEL_FUEL:
        Denotes the diesel fuel component of the system.

    - DIESEL_WATER_HEATER:
        Denotes the diesel water heater component of the system.

    - ELECTRIC_WATER_HEATER:
        Denotes the electric water heater component of the system.

    - GENERAL:
        Denotes impacts generally associated with the system but not a specific
        component.

    - GRID:
        Denotes the grid component of the system.

    - HEAT_EXCHANGER:
        Denotes the heat exchanger component of the system.

    - HOT_WATER_TANK:
        Denotes the hot-water tank component of the system.

    - HOUSEHOLDS:
        Denotes households.

    - INVERTER:
        Denotes the inverter component of the system.

    - KEROSENE:
        Denotes the kerosene component of the system.

    - MISC:
        Denotes misc. aspects of the system.

    - PV:
        Denotes the PV component of the system.

    - PV_T:
        Denotes the PV-T component of the system.

    - STORAGE:
        Denotes the storage component of the system.

    - TRANSMITTER:
        Denotes a component of the system responsible for transmitting one resource type
        at the expense of another.

    """

    BOS = "bos"
    BUFFER_TANK = "buffer_tank"
    CLEAN_WATER_TANK = "clean_water_tank"
    CONVENTIONAL_SOURCE = "conventional_source"
    CONVERTER = "converter"
    DIESEL = "diesel_generator"
    DIESEL_FUEL = "diesel_fuel"
    DIESEL_WATER_HEATER = "diesel_water_heater"
    ELECTRIC_WATER_HEATER = "electric_water_heater"
    GENERAL = "general"
    GRID = "grid"
    HEAT_EXCHANGER = EXCHANGER
    HOUSEHOLDS = "households"
    HOT_WATER_TANK = "hot_water_tank"
    INVERTER = "inverter"
    KEROSENE = "kerosene"
    MISC = "misc"
    PV = "pv"
    PV_T = "pv_t"
    STORAGE = "storage"
    TRANSMITTER = "transmitter"


class WasteProduct(enum.Enum):
    """
    Used to keep track of waste products generated by the system.

    - BRINE:
        Denotes the brine that is produced as a by-product of desalination.

    """

    BRINE = "brine"
